import os
import pickle

import numpy as np
import pandas as pd
import pytest
import torch

from scvi.data import synthetic_iid
from scvi.model import SCANVI, SCVI
from scvi.utils import attrdict


def test_saving_and_loading(save_path):
    def legacy_save(
        model,
        dir_path,
        prefix=None,
        overwrite=False,
        save_anndata=False,
        **anndata_write_kwargs,
    ):
        if not os.path.exists(dir_path) or overwrite:
            os.makedirs(dir_path, exist_ok=overwrite)
        else:
            raise ValueError(
                f"{dir_path} already exists. Please provide an unexisting directory for saving."
            )

        file_name_prefix = prefix or ""

        if save_anndata:
            model.adata.write(
                os.path.join(dir_path, f"{file_name_prefix}adata.h5ad"),
                **anndata_write_kwargs,
            )

        model_save_path = os.path.join(dir_path, f"{file_name_prefix}model_params.pt")
        attr_save_path = os.path.join(dir_path, f"{file_name_prefix}attr.pkl")
        varnames_save_path = os.path.join(dir_path, f"{file_name_prefix}var_names.csv")

        torch.save(model.module.state_dict(), model_save_path)

        var_names = model.adata.var_names.astype(str)
        var_names = var_names.to_numpy()
        np.savetxt(varnames_save_path, var_names, fmt="%s")

        # get all the user attributes
        user_attributes = model._get_user_attributes()
        # only save the public attributes with _ at the very end
        user_attributes = {a[0]: a[1] for a in user_attributes if a[0][-1] == "_"}

        with open(attr_save_path, "wb") as f:
            pickle.dump(user_attributes, f)

    save_path = os.path.join(save_path, "tmp")
    adata = synthetic_iid()

    prefix = "SCANVI_"
    SCANVI.setup_anndata(adata, "labels", "label_0", batch_key="batch")
    model = SCANVI(adata)
    model.train(max_epochs=1, train_size=0.5)
    p1 = model.predict()
    model.save(save_path, overwrite=True, save_anndata=True, prefix=prefix)
    model.view_setup_args(save_path, prefix=prefix)
    model = SCANVI.load(save_path, prefix=prefix)
    model.get_latent_representation()
    tmp_adata = synthetic_iid(n_genes=200)
    with pytest.raises(ValueError):
        SCANVI.load(save_path, adata=tmp_adata, prefix=prefix)
    model = SCANVI.load(save_path, adata=adata, prefix=prefix)
    assert "batch" in model.adata_manager.data_registry
    assert model.adata_manager.data_registry.batch == attrdict(
        {"attr_name": "obs", "attr_key": "_scvi_batch"}
    )

    p2 = model.predict()
    np.testing.assert_array_equal(p1, p2)
    assert model.is_trained is True

    # Test legacy loading
    legacy_save_path = os.path.join(save_path, "legacy/")
    legacy_save(model, legacy_save_path, overwrite=True, save_anndata=True, prefix=prefix)
    with pytest.raises(ValueError):
        SCANVI.load(legacy_save_path, adata=adata, prefix=prefix)
    SCANVI.convert_legacy_save(legacy_save_path, legacy_save_path, overwrite=True, prefix=prefix)
    m = SCANVI.load(legacy_save_path, adata=adata, prefix=prefix)
    m.train(1)


def test_scanvi():
    adata = synthetic_iid()
    SCANVI.setup_anndata(
        adata,
        "labels",
        "label_0",
        batch_key="batch",
    )
    model = SCANVI(adata, n_latent=10)
    assert len(model._labeled_indices) == sum(adata.obs["labels"] != "label_0")
    assert len(model._unlabeled_indices) == sum(adata.obs["labels"] == "label_0")
    model.train(1, train_size=0.5, check_val_every_n_epoch=1)
    logged_keys = model.history.keys()
    assert "elbo_validation" in logged_keys
    assert "reconstruction_loss_validation" in logged_keys
    assert "kl_local_validation" in logged_keys
    assert "elbo_train" in logged_keys
    assert "reconstruction_loss_train" in logged_keys
    assert "kl_local_train" in logged_keys
    assert "validation_classification_loss" in logged_keys
    assert "validation_accuracy" in logged_keys
    assert "validation_f1_score" in logged_keys
    assert "validation_calibration_error" in logged_keys
    adata2 = synthetic_iid()
    predictions = model.predict(adata2, indices=[1, 2, 3])
    assert len(predictions) == 3
    model.predict()
    df = model.predict(adata2, soft=True)
    assert isinstance(df, pd.DataFrame)
    model.predict(adata2, soft=True, indices=[1, 2, 3])
    model.get_normalized_expression(adata2)
    model.differential_expression(groupby="labels", group1="label_1")
    model.differential_expression(groupby="labels", group1="label_1", group2="label_2")

    # test that all data labeled runs
    unknown_label = "asdf"
    a = synthetic_iid()
    SCANVI.setup_anndata(
        a,
        "labels",
        unknown_label,
        batch_key="batch",
    )
    m = SCANVI(a)
    m.train(1)

    # test mix of labeled and unlabeled data
    unknown_label = "label_0"
    a = synthetic_iid()
    SCANVI.setup_anndata(
        a,
        "labels",
        unknown_label,
        batch_key="batch",
    )
    m = SCANVI(a)
    m.train(1, train_size=0.9)

    # test from_scvi_model
    a = synthetic_iid()
    SCVI.setup_anndata(
        a,
        batch_key="batch",
    )
    m = SCVI(a, use_observed_lib_size=False)
    a2 = synthetic_iid()
    scanvi_model = SCANVI.from_scvi_model(m, "label_0", labels_key="labels", adata=a2)
    with pytest.raises(ValueError):
        scanvi_model = SCANVI.from_scvi_model(m, "label_0", labels_key=None, adata=a2)

    # make sure the state_dicts are different objects for the two models
    assert scanvi_model.module.state_dict() is not m.module.state_dict()
    scanvi_pxr = scanvi_model.module.state_dict().get("px_r", None)
    scvi_pxr = m.module.state_dict().get("px_r", None)
    assert scanvi_pxr is not None
    assert scvi_pxr is not None
    assert scanvi_pxr is not scvi_pxr
    scanvi_model.train(1)

    # Test without label groups
    scanvi_model = SCANVI.from_scvi_model(
        m, "label_0", labels_key="labels", use_labels_groups=False
    )
    scanvi_model.train(1)

    # test from_scvi_model with size_factor
    a = synthetic_iid()
    a.obs["size_factor"] = np.random.randint(1, 5, size=(a.shape[0],))
    SCVI.setup_anndata(a, batch_key="batch", labels_key="labels", size_factor_key="size_factor")
    m = SCVI(a, use_observed_lib_size=False)
    a2 = synthetic_iid()
    a2.obs["size_factor"] = np.random.randint(1, 5, size=(a2.shape[0],))
    scanvi_model = SCANVI.from_scvi_model(m, "label_0", adata=a2)
    scanvi_model.train(1)


def test_scanvi_with_external_indices():
    adata = synthetic_iid()
    SCANVI.setup_anndata(
        adata,
        "labels",
        "label_0",
        batch_key="batch",
    )
    model = SCANVI(adata, n_latent=10)
    assert len(model._labeled_indices) == sum(adata.obs["labels"] != "label_0")
    assert len(model._unlabeled_indices) == sum(adata.obs["labels"] == "label_0")
    # in this case we will make a stratified version of indexing
    from sklearn.model_selection import train_test_split

    train_ind, valid_ind = train_test_split(
        adata.obs.batch.index.astype(int), test_size=0.6, stratify=adata.obs.batch
    )
    test_ind, valid_ind = train_test_split(
        valid_ind, test_size=0.5, stratify=adata.obs.batch[valid_ind]
    )
    model.train(
        1,
        check_val_every_n_epoch=1,
        datasplitter_kwargs={
            "external_indexing": [np.array(train_ind), np.array(valid_ind), np.array(test_ind)]
        },
    )
    logged_keys = model.history.keys()
    assert "elbo_validation" in logged_keys
    assert "reconstruction_loss_validation" in logged_keys
    assert "kl_local_validation" in logged_keys
    assert "elbo_train" in logged_keys
    assert "reconstruction_loss_train" in logged_keys
    assert "kl_local_train" in logged_keys
    assert "validation_classification_loss" in logged_keys
    assert "validation_accuracy" in logged_keys
    assert "validation_f1_score" in logged_keys
    assert "validation_calibration_error" in logged_keys
    adata2 = synthetic_iid()
    predictions = model.predict(adata2, indices=[1, 2, 3])
    assert len(predictions) == 3
    model.predict()
    df = model.predict(adata2, soft=True)
    assert isinstance(df, pd.DataFrame)
    model.predict(adata2, soft=True, indices=[1, 2, 3])
    model.get_normalized_expression(adata2)
    model.differential_expression(groupby="labels", group1="label_1")
    model.differential_expression(groupby="labels", group1="label_1", group2="label_2")

    # test that all data labeled runs
    unknown_label = "asdf"
    a = synthetic_iid()
    SCANVI.setup_anndata(
        a,
        "labels",
        unknown_label,
        batch_key="batch",
    )
    m = SCANVI(a)
    m.train(1)

    # test mix of labeled and unlabeled data
    unknown_label = "label_0"
    a = synthetic_iid()
    SCANVI.setup_anndata(
        a,
        "labels",
        unknown_label,
        batch_key="batch",
    )
    m = SCANVI(a)
    m.train(1, train_size=0.9)

    # test from_scvi_model
    a = synthetic_iid()
    SCVI.setup_anndata(
        a,
        batch_key="batch",
    )
    m = SCVI(a, use_observed_lib_size=False)
    a2 = synthetic_iid()
    scanvi_model = SCANVI.from_scvi_model(m, "label_0", labels_key="labels", adata=a2)
    with pytest.raises(ValueError):
        scanvi_model = SCANVI.from_scvi_model(m, "label_0", labels_key=None, adata=a2)

    # make sure the state_dicts are different objects for the two models
    assert scanvi_model.module.state_dict() is not m.module.state_dict()
    scanvi_pxr = scanvi_model.module.state_dict().get("px_r", None)
    scvi_pxr = m.module.state_dict().get("px_r", None)
    assert scanvi_pxr is not None
    assert scvi_pxr is not None
    assert scanvi_pxr is not scvi_pxr
    scanvi_model.train(1)

    # Test without label groups
    scanvi_model = SCANVI.from_scvi_model(
        m, "label_0", labels_key="labels", use_labels_groups=False
    )
    scanvi_model.train(1)

    # test from_scvi_model with size_factor
    a = synthetic_iid()
    a.obs["size_factor"] = np.random.randint(1, 5, size=(a.shape[0],))
    SCVI.setup_anndata(a, batch_key="batch", labels_key="labels", size_factor_key="size_factor")
    m = SCVI(a, use_observed_lib_size=False)
    a2 = synthetic_iid()
    a2.obs["size_factor"] = np.random.randint(1, 5, size=(a2.shape[0],))
    scanvi_model = SCANVI.from_scvi_model(m, "label_0", adata=a2)
    scanvi_model.train(1)


def test_scanvi_predict_use_posterior_mean():
    adata = synthetic_iid()
    SCANVI.setup_anndata(adata, labels_key="labels", unlabeled_category="label_0")

    model = SCANVI(adata)
    model.train(max_epochs=1)

    _ = model.predict(use_posterior_mean=True)
    _ = model.predict(use_posterior_mean=False)


def test_linear_classifier_scanvi(n_latent: int = 10, n_labels: int = 5):
    adata = synthetic_iid(n_labels=n_labels)
    SCANVI.setup_anndata(
        adata,
        "labels",
        "label_0",
        batch_key="batch",
    )
    model = SCANVI(adata, linear_classifier=True, n_latent=n_latent)

    assert isinstance(model.module.classifier.classifier[0], torch.nn.Linear)
    assert model.module.classifier.classifier[0].in_features == n_latent
    assert model.module.classifier.classifier[0].out_features == n_labels - 1

    model.train(max_epochs=1)


def test_multiple_covariates_scanvi():
    adata = synthetic_iid()
    adata.obs["cont1"] = np.random.normal(size=(adata.shape[0],))
    adata.obs["cont2"] = np.random.normal(size=(adata.shape[0],))
    adata.obs["cat1"] = np.random.randint(0, 5, size=(adata.shape[0],))
    adata.obs["cat2"] = np.random.randint(0, 5, size=(adata.shape[0],))

    SCANVI.setup_anndata(
        adata,
        "labels",
        "Unknown",
        batch_key="batch",
        continuous_covariate_keys=["cont1", "cont2"],
        categorical_covariate_keys=["cat1", "cat2"],
    )
    m = SCANVI(adata)
    m.train(1)
    m.get_latent_representation()
    m.get_elbo()
    m.get_marginal_ll(n_mc_samples=3)
    m.get_marginal_ll(adata, return_mean=True, n_mc_samples=6, n_mc_samples_per_pass=1)
    m.get_marginal_ll(adata, return_mean=True, n_mc_samples=6, n_mc_samples_per_pass=6)
    m.differential_expression(
        idx1=np.arange(50), idx2=51 + np.arange(50), mode="vanilla", weights="uniform"
    )
    m.differential_expression(
        idx1=np.arange(50),
        idx2=51 + np.arange(50),
        mode="vanilla",
        weights="importance",
        importance_weighting_kwargs={"n_mc_samples": 10, "n_mc_samples_per_pass": 1},
    )
    m.differential_expression(
        idx1=np.arange(50),
        idx2=51 + np.arange(50),
        mode="vanilla",
        weights="importance",
        importance_weighting_kwargs={"n_mc_samples": 10, "n_mc_samples_per_pass": 10},
    )
    m.get_reconstruction_error()
    m.get_normalized_expression(n_samples=1)
    m.get_normalized_expression(n_samples=2)


def test_multiple_encoded_covariates_scanvi():
    adata = synthetic_iid()
    adata.obs["cont1"] = np.random.normal(size=(adata.shape[0],))
    adata.obs["cont2"] = np.random.normal(size=(adata.shape[0],))
    adata.obs["cat1"] = np.random.randint(0, 5, size=(adata.shape[0],))
    adata.obs["cat2"] = np.random.randint(0, 5, size=(adata.shape[0],))

    SCANVI.setup_anndata(
        adata,
        "labels",
        "Unknown",
        batch_key="batch",
        continuous_covariate_keys=["cont1", "cont2"],
        categorical_covariate_keys=["cat1", "cat2"],
    )
    m = SCANVI(adata, encode_covariates=True)
    m.train(1)


def test_scanvi_online_update(save_path):
    # ref has semi-observed labels
    n_latent = 5
    adata1 = synthetic_iid()
    new_labels = adata1.obs.labels.to_numpy()
    new_labels[0] = "Unknown"
    adata1.obs["labels"] = pd.Categorical(new_labels)
    adata1.obs["cont1"] = np.random.normal(size=(adata1.shape[0],))
    adata1.obs["cont2"] = np.random.normal(size=(adata1.shape[0],))

    SCANVI.setup_anndata(
        adata1,
        "labels",
        "Unknown",
        batch_key="batch",
        continuous_covariate_keys=["cont1", "cont2"],
    )
    model = SCANVI(
        adata1,
        n_latent=n_latent,
        encode_covariates=True,
    )
    model.train(max_epochs=1, check_val_every_n_epoch=1)
    dir_path = os.path.join(save_path, "saved_model/")
    model.save(dir_path, overwrite=True)

    # query has all missing labels
    adata2 = synthetic_iid()
    adata2.obs["batch"] = adata2.obs.batch.cat.rename_categories(["batch_2", "batch_3"])
    adata2.obs["labels"] = "Unknown"
    adata2.obs["cont1"] = np.random.normal(size=(adata2.shape[0],))
    adata2.obs["cont2"] = np.random.normal(size=(adata2.shape[0],))

    model = SCANVI.load_query_data(adata2, dir_path, freeze_batchnorm_encoder=True)
    model.train(max_epochs=1)
    model.get_latent_representation()
    model.predict()

    # query has all missing labels and no labels key
    del adata2.obs["labels"]

    model = SCANVI.load_query_data(adata2, dir_path, freeze_batchnorm_encoder=True)
    model.train(max_epochs=1)
    model.get_latent_representation()
    model.predict()

    # query has no missing labels
    adata2 = synthetic_iid()
    adata2.obs["batch"] = adata2.obs.batch.cat.rename_categories(["batch_2", "batch_3"])
    adata2.obs["cont1"] = np.random.normal(size=(adata2.shape[0],))
    adata2.obs["cont2"] = np.random.normal(size=(adata2.shape[0],))

    model = SCANVI.load_query_data(adata2, dir_path, freeze_batchnorm_encoder=True)
    model.train(max_epochs=1)
    model.get_latent_representation()
    model.predict()

    # Test on extra categoricals as well
    adata1 = synthetic_iid()
    new_labels = adata1.obs.labels.to_numpy()
    new_labels[0] = "Unknown"
    adata1.obs["labels"] = pd.Categorical(new_labels)
    adata1.obs["cont1"] = np.random.normal(size=(adata1.shape[0],))
    adata1.obs["cont2"] = np.random.normal(size=(adata1.shape[0],))
    adata1.obs["cat1"] = np.random.randint(0, 5, size=(adata1.shape[0],))
    adata1.obs["cat2"] = np.random.randint(0, 5, size=(adata1.shape[0],))
    SCANVI.setup_anndata(
        adata1,
        "labels",
        "Unknown",
        batch_key="batch",
        continuous_covariate_keys=["cont1", "cont2"],
        categorical_covariate_keys=["cat1", "cat2"],
    )
    model = SCANVI(
        adata1,
        n_latent=n_latent,
        encode_covariates=True,
    )
    model.train(max_epochs=1, check_val_every_n_epoch=1)
    dir_path = os.path.join(save_path, "saved_model/")
    model.save(dir_path, overwrite=True)

    adata2 = synthetic_iid()
    adata2.obs["batch"] = adata2.obs.batch.cat.rename_categories(["batch_2", "batch_3"])
    adata2.obs["labels"] = "Unknown"
    adata2.obs["cont1"] = np.random.normal(size=(adata2.shape[0],))
    adata2.obs["cont2"] = np.random.normal(size=(adata2.shape[0],))
    adata2.obs["cat1"] = np.random.randint(0, 5, size=(adata2.shape[0],))
    adata2.obs["cat2"] = np.random.randint(0, 5, size=(adata2.shape[0],))
    SCANVI.load_query_data(adata2, dir_path, freeze_batchnorm_encoder=True)

    # ref has fully-observed labels
    n_latent = 5
    adata1 = synthetic_iid()
    new_labels = adata1.obs.labels.to_numpy()
    adata1.obs["labels"] = pd.Categorical(new_labels)
    SCANVI.setup_anndata(adata1, "labels", "Unknown", batch_key="batch")
    model = SCANVI(adata1, n_latent=n_latent, encode_covariates=True)
    model.train(max_epochs=1, check_val_every_n_epoch=1)
    dir_path = os.path.join(save_path, "saved_model/")
    model.save(dir_path, overwrite=True)

    # query has one new label
    adata2 = synthetic_iid()
    adata2.obs["batch"] = adata2.obs.batch.cat.rename_categories(["batch_2", "batch_3"])
    new_labels = adata2.obs.labels.to_numpy()
    new_labels[0] = "Unknown"
    adata2.obs["labels"] = pd.Categorical(new_labels)

    model2 = SCANVI.load_query_data(adata2, dir_path, freeze_batchnorm_encoder=True)
    model2._unlabeled_indices = np.arange(adata2.n_obs)
    model2._labeled_indices = []
    model2.train(max_epochs=1, plan_kwargs={"weight_decay": 0.0})
    model2.get_latent_representation()
    model2.predict()

    # test classifier frozen
    class_query_weight = (
        model2.module.classifier.classifier[0].fc_layers[0][0].weight.detach().cpu().numpy()
    )
    class_ref_weight = (
        model.module.classifier.classifier[0].fc_layers[0][0].weight.detach().cpu().numpy()
    )
    # weight decay makes difference
    np.testing.assert_allclose(class_query_weight, class_ref_weight, atol=1e-07)

    # test classifier unfrozen
    model2 = SCANVI.load_query_data(adata2, dir_path, freeze_classifier=False)
    model2._unlabeled_indices = np.arange(adata2.n_obs)
    model2._labeled_indices = []
    model2.train(max_epochs=1)
    class_query_weight = (
        model2.module.classifier.classifier[0].fc_layers[0][0].weight.detach().cpu().numpy()
    )
    class_ref_weight = (
        model.module.classifier.classifier[0].fc_layers[0][0].weight.detach().cpu().numpy()
    )
    with pytest.raises(AssertionError):
        np.testing.assert_allclose(class_query_weight, class_ref_weight, atol=1e-07)

    # test saving and loading of online scanvi
    a = synthetic_iid()
    ref = a[a.obs["labels"] != "label_2"].copy()  # only has labels 0 and 1
    SCANVI.setup_anndata(ref, "labels", "label_2", batch_key="batch")
    m = SCANVI(ref)
    m.train(max_epochs=1)
    m.save(save_path, overwrite=True)
    query = a[a.obs["labels"] != "label_0"].copy()
    query = synthetic_iid()  # has labels 0 and 2. 2 is unknown
    m_q = SCANVI.load_query_data(query, save_path)
    m_q.save(save_path, overwrite=True)
    m_q = SCANVI.load(save_path, adata=query)
    m_q.predict()
    m_q.get_elbo()


def test_scanvi_logits_backwards_compat(save_path: str):
    """Check that we can replicate pre-fix behavior with `logits=False`."""
    adata = synthetic_iid()
    SCANVI.setup_anndata(adata, labels_key="labels", unlabeled_category="label_0")

    model = SCANVI(adata, classifier_parameters={"logits": False})
    model.train(max_epochs=1)

    model_path = os.path.join(save_path, "scanvi_logits_model")
    model.save(model_path, overwrite=True)
    del model

    model = SCANVI.load(model_path, adata)
    assert not model.module.classifier.logits
    assert isinstance(model.module.classifier.classifier[-1], torch.nn.Softmax)


def test_scanvi_pre_logits_fix_load(save_path: str):
    """See #2310. Check old model saves use the old behavior."""
    model_path = "tests/test_data/pre_logits_fix_scanvi"
    model = SCANVI.load(model_path)

    def check_no_logits_and_softmax(model: SCANVI):
        assert not model.module.classifier.logits
        assert isinstance(model.module.classifier.classifier[-1], torch.nn.Softmax)

    check_no_logits_and_softmax(model)

    resave_model_path = os.path.join(save_path, "pre_logits_fix_scanvi")
    model.save(resave_model_path, overwrite=True)
    adata = model.adata
    del model

    model = SCANVI.load(resave_model_path, adata)
    check_no_logits_and_softmax(model)



from __future__ import annotations

from typing import TYPE_CHECKING

import torch

from scvi import REGISTRY_KEYS
from scvi.module.base import EmbeddingModuleMixin

if TYPE_CHECKING:
    import numpy as np
    from anndata import AnnData


class EmbeddingMixin:
    """``EXPERIMENTAL`` Mixin class for initializing and using embeddings in a model.

    Must be used with a module that inherits from :class:`~scvi.module.base.EmbeddingModuleMixin`.

    Notes
    -----
    Lifecycle: experimental in v1.2.
    """

    @torch.inference_mode()
    def get_batch_representation(
        self,
        adata: AnnData | None = None,
        indices: list[int] | None = None,
        batch_size: int | None = None,
    ) -> np.ndarray:
        """Get the batch representation for a given set of indices."""
        if not isinstance(self.module, EmbeddingModuleMixin):
            raise ValueError("The current `module` must inherit from `EmbeddingModuleMixin`.")

        adata = self._validate_anndata(adata)
        dataloader = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)

        key = REGISTRY_KEYS.BATCH_KEY
        tensors = [self.module.compute_embedding(key, tensors[key]) for tensors in dataloader]
        return torch.cat(tensors).detach().cpu().numpy()



from ._arraylike_field import (
    CategoricalJointObsField,
    CategoricalJointVarField,
    MuDataCategoricalJointObsField,
    MuDataCategoricalJointVarField,
    MuDataNumericalJointObsField,
    MuDataNumericalJointVarField,
    MuDataObsmField,
    MuDataVarmField,
    NumericalJointObsField,
    NumericalJointVarField,
    ObsmField,
    VarmField,
)
from ._base_field import AnnDataField, BaseAnnDataField
from ._dataframe_field import (
    CategoricalObsField,
    CategoricalVarField,
    MuDataCategoricalObsField,
    MuDataCategoricalVarField,
    MuDataNumericalObsField,
    MuDataNumericalVarField,
    NumericalObsField,
    NumericalVarField,
)
from ._layer_field import LayerField, MuDataLayerField
from ._mudata import BaseMuDataWrapperClass, MuDataWrapper
from ._protein import MuDataProteinLayerField, ProteinObsmField
from ._scanvi import LabelsWithUnlabeledObsField
from ._uns_field import StringUnsField

__all__ = [
    "BaseAnnDataField",
    "BaseMuDataWrapperClass",
    "MuDataWrapper",
    "AnnDataField",
    "LayerField",
    "MuDataLayerField",
    "MuDataProteinLayerField",
    "NumericalObsField",
    "NumericalVarField",
    "MuDataNumericalObsField",
    "MuDataNumericalVarField",
    "CategoricalObsField",
    "CategoricalVarField",
    "MuDataCategoricalObsField",
    "MuDataCategoricalVarField",
    "ObsmField",
    "VarmField",
    "MuDataObsmField",
    "MuDataVarmField",
    "NumericalJointObsField",
    "NumericalJointVarField",
    "MuDataNumericalJointObsField",
    "MuDataNumericalJointVarField",
    "CategoricalJointObsField",
    "CategoricalJointVarField",
    "MuDataCategoricalJointObsField",
    "MuDataCategoricalJointVarField",
    "ProteinObsmField",
    "LabelsWithUnlabeledObsField",
    "StringUnsField",
]



from __future__ import annotations

import json
import warnings
from typing import TYPE_CHECKING, Literal

import numpy as np
import pandas as pd
from anndata import AnnData
from mudata import MuData
from scipy.sparse import issparse
from scipy.stats import pearsonr, spearmanr
from sklearn.metrics import (
    average_precision_score,
    precision_recall_fscore_support,
    roc_auc_score,
)
from sparse import GCXS, SparseArray
from xarray import DataArray, Dataset

from scvi.utils import dependencies

from ._constants import (
    DATA_VAR_RAW,
    DEFAULT_DE_N_TOP_GENES_OVERLAP,
    DEFAULT_DE_P_VAL_THRESHOLD,
    METRIC_CALIBRATION,
    METRIC_CV_CELL,
    METRIC_CV_GENE,
    METRIC_DIFF_EXP,
    METRIC_ZERO_FRACTION,
    UNS_NAME_RGG_PPC,
    UNS_NAME_RGG_RAW,
)

if TYPE_CHECKING:
    from scvi._types import AnnOrMuData
    from scvi.model.base import BaseModelClass

Dims = Literal["cells", "features"]


def _make_dataset_dense(dataset: Dataset) -> Dataset:
    """Make a dataset dense, converting sparse arrays to dense arrays."""
    dataset = dataset.map(lambda x: x.data.todense() if isinstance(x.data, SparseArray) else x)
    return dataset


class PosteriorPredictiveCheck:
    """
    ``EXPERIMENTAL`` Posterior predictive checks for comparing scRNA-seq generative models.

    Parameters
    ----------
    adata
        :class:`~AnnOrMudata` object with raw counts in either ``adata.X`` or ``adata.layers``.
    models_dict
        Dictionary of models to compare.
    count_layer_key
        Key in ``adata.layers`` to use as raw counts. If ``None``, defaults to ``adata.X``.
    n_samples
        Number of posterior predictive samples to generate.
    indices
        Indices of observations in ``adata`` to subset to before generating posterior predictive
        samples and computing metrics. If ``None``, defaults to all observations in ``adata``.
    modality
        Modality to use for posterior predictive samples. Needs to be defined if using MuData
    """

    def __init__(
        self,
        adata: AnnOrMuData,
        models_dict: dict[str, BaseModelClass],
        count_layer_key: str | None = None,
        n_samples: int = 10,
        indices: list | None = None,
        modality: str | None = None,
    ):
        if indices is not None:
            adata = adata[indices]
        self.count_layer_key = count_layer_key
        self.modality = modality
        if isinstance(adata, MuData):
            assert modality is not None, "Modality must be defined for MuData."
            self.adata = adata[modality]
            raw_counts = (
                self.adata.layers[count_layer_key] if count_layer_key is not None else self.adata.X
            )
        else:
            self.adata = adata
            raw_counts = adata.layers[count_layer_key] if count_layer_key is not None else adata.X
        # Compressed axis is rows, like csr
        if isinstance(raw_counts, np.ndarray):
            self.raw_counts = GCXS.from_numpy(raw_counts, compressed_axes=(0,))
        elif issparse(raw_counts):
            self.raw_counts = GCXS.from_scipy_sparse(raw_counts).change_compressed_axes((0,))
        else:
            raise ValueError("raw_counts must be a numpy array or scipy sparse matrix")
        self.samples_dataset = None
        self.n_samples = n_samples
        self.models = models_dict
        self.metrics = {}

        self._store_posterior_predictive_samples(indices=indices)

    def __repr__(self) -> str:
        return (
            f"--- Posterior Predictive Checks ---\n"
            f"n_samples = {self.n_samples}\n"
            f"raw_counts shape = {self.raw_counts.shape}\n"
            f"models: {list(self.models.keys())}\n"
            f"metrics: \n{self._metrics_repr()}"
        )

    def _metrics_repr(self) -> str:
        def custom_handle_unserializable(o):
            if isinstance(o, AnnData):
                return f"AnnData object with n_obs={o.n_obs}, n_vars={o.n_vars}"
            elif isinstance(o, pd.DataFrame):
                s = f"Pandas DataFrame with shape={o.shape}, "
                n_cols = 5
                if len(o.columns) > n_cols:
                    return s + f"first {n_cols} columns={o.columns[:n_cols].to_list()}"
                return s + f"columns={o.columns.to_list()}"
            elif isinstance(o, pd.Series):
                return f"Pandas Series with n_rows={len(o)}"
            return f"ERROR unserializable type: {type(o)}"

        return json.dumps(self.metrics, indent=4, default=custom_handle_unserializable)

    def _store_posterior_predictive_samples(
        self,
        batch_size: int = 32,
        indices: list[int] | None = None,
    ):
        """
        Store posterior predictive samples for each model.

        Parameters
        ----------
        models_dict
            Dictionary of models to store posterior predictive samples for.
        batch_size
            Batch size for generating posterior predictive samples.
        indices
            Indices to generate posterior predictive samples for.
        """
        self.batch_size = batch_size

        samples_dict = {}
        for m, model in self.models.items():
            pp_counts = model.posterior_predictive_sample(
                model.adata,
                n_samples=self.n_samples,
                batch_size=self.batch_size,
                indices=indices,
            )
            if isinstance(pp_counts, dict):
                pp_counts = pp_counts[self.modality]
            samples_dict[m] = DataArray(
                data=pp_counts,
                coords={
                    "cells": list(self.adata.obs_names),
                    "features": list(self.adata.var_names),
                    "samples": np.arange(self.n_samples),
                },
            )
        samples_dict[DATA_VAR_RAW] = DataArray(
            data=self.raw_counts,
            coords={"cells": list(self.adata.obs_names), "features": list(self.adata.var_names)},
        )
        self.samples_dataset = Dataset(samples_dict)

    def coefficient_of_variation(self, dim: Dims = "cells") -> None:
        """
        Calculate the coefficient of variation (CV) for each model and the raw counts.

        The CV is computed over the cells or features dimension per sample. The mean CV is then
        computed over all samples.

        Parameters
        ----------
        dim
            Dimension to compute CV over.
        """
        identifier = METRIC_CV_CELL if dim == "features" else METRIC_CV_GENE
        mean = self.samples_dataset.mean(dim=dim, skipna=False)
        # we use a trick to compute the std to speed it up: std = E[X^2] - E[X]^2
        # a square followed by a sqrt is ok here because this is counts data (no negative values)
        self.samples_dataset = np.square(self.samples_dataset)
        std = np.sqrt(self.samples_dataset.mean(dim=dim, skipna=False) - np.square(mean))
        self.samples_dataset = np.sqrt(self.samples_dataset)
        # now compute the CV
        cv = std / mean
        # It's ok to make things dense here
        cv = _make_dataset_dense(cv)
        cv_mean = cv.mean(dim="samples", skipna=True)
        cv_mean[DATA_VAR_RAW].data = np.nan_to_num(cv_mean[DATA_VAR_RAW].data)
        self.metrics[identifier] = cv_mean.to_dataframe()

    def zero_fraction(self) -> None:
        """Fraction of zeros in raw counts for a specific gene"""
        pp_samples = self.samples_dataset
        mean = (pp_samples != 0).mean(dim="cells", skipna=False).mean(dim="samples", skipna=False)
        mean = _make_dataset_dense(mean)
        self.metrics[METRIC_ZERO_FRACTION] = mean.to_dataframe()

    def calibration_error(self, confidence_intervals: list[float] | float = None) -> None:
        """Calibration error for each observed count.

        For a series of credible intervals of the samples, the fraction of observed counts that
        fall within the credible interval is computed. The calibration error is then the squared
        difference between the observed fraction and the true interval width.

        For this metric, lower is better.

        Parameters
        ----------
        confidence_intervals
            List of confidence intervals to compute calibration error for.
            E.g., [0.01, 0.02, 0.98, 0.99]

        Notes
        -----
        This does not work on sparse data and can cause large memory usage.
        """
        if confidence_intervals is None:
            ps = [2.5, 5, 7.5, 10, 12.5, 15, 17.5, 82.5, 85, 87.5, 90, 92.5, 95, 97.5]
            ps = [p / 100 for p in ps]
        else:
            if len(confidence_intervals) % 2 != 0:
                raise ValueError("Confidence intervals must be even")
            ps = confidence_intervals
        pp_samples = self.samples_dataset
        # TODO: Reimplement to work on sparse data
        pp_samples = _make_dataset_dense(pp_samples)
        # results in (quantiles, cells, features)
        quants = pp_samples.quantile(q=ps, dim="samples", skipna=False)
        credible_interval_indices = [(i, len(ps) - (i + 1)) for i in range(len(ps) // 2)]

        model_cal = {}
        for model in pp_samples.data_vars:
            if model == DATA_VAR_RAW:
                continue
            cal_error_features = 0
            for interval in credible_interval_indices:
                start = interval[0]
                end = interval[1]
                true_width = ps[end] - ps[start]
                greater_than = (quants[DATA_VAR_RAW] >= quants.model1.isel(quantile=start)).data
                less_than = (quants[DATA_VAR_RAW] <= quants.model1.isel(quantile=end)).data
                # Logical and
                ci = greater_than * less_than
                pci_features = ci.mean()
                cal_error_features += (pci_features - true_width) ** 2
            model_cal[model] = {
                "features": cal_error_features,
            }
        self.metrics[METRIC_CALIBRATION] = pd.DataFrame.from_dict(model_cal)

    @dependencies("scanpy")
    def differential_expression(
        self,
        de_groupby: str | None = None,
        de_method: str = "t-test",
        n_samples: int = 1,
        cell_scale_factor: float = 1e4,
        p_val_thresh: float = DEFAULT_DE_P_VAL_THRESHOLD,
        n_top_genes_fallback: int = DEFAULT_DE_N_TOP_GENES_OVERLAP,
    ):
        """
        Compute differential expression (DE) metrics.

        If n_samples > 1, all metrics are averaged over a posterior predictive dataset.

        Parameters
        ----------
        de_groupby
            The column name in `adata_obs_raw` that contains the groupby information.
        de_method
            The DE method to use. See :meth:`~scanpy.tl.rank_genes_groups` for more details.
        n_samples
            The number of posterior predictive samples to use for the DE analysis.
        cell_scale_factor
            The cell scale factor to use for normalization before DE.
        p_val_thresh
            The p-value threshold to use for the DE analysis.
        n_top_genes_fallback
            The number of top genes to use for the DE analysis if the number of genes
            with a p-value < p_val_thresh is zero.
        """
        if 10 * n_top_genes_fallback > self.adata.n_vars:
            warnings.warn(
                f"n_top_genes_fallback={n_top_genes_fallback} is greater than 10% of n_vars"
                f" {self.adata.n_vars} in the dataset. Setting it to 10% of n_vars.",
                UserWarning,
                stacklevel=2,
            )
            n_top_genes_fallback = int(0.1 * self.adata.n_vars)

        import scanpy as sc

        if n_samples > self.n_samples:
            raise ValueError(
                f"n_samples={n_samples} is greater than the number of samples already recorded "
                f"({self.n_samples})"
            )
        # run DE with the raw counts
        adata_de = AnnData(
            X=self.raw_counts.to_scipy_sparse().tocsr().copy(),
            obs=self.adata.obs,
            var=self.adata.var,
        )
        sc.pp.normalize_total(adata_de, target_sum=cell_scale_factor)
        sc.pp.log1p(adata_de)
        if de_groupby is None:
            sc.tl.pca(adata_de)
            sc.pp.neighbors(adata_de)
            sc.tl.leiden(adata_de, key_added="leiden_scvi_criticism")
            de_groupby = "leiden_scvi_criticism"
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
            sc.tl.rank_genes_groups(
                adata_de,
                de_groupby,
                use_raw=False,
                method=de_method,
                key_added=UNS_NAME_RGG_RAW,
            )

        # get posterior predictive samples from the model (aka approx counts)
        pp_samples = self.samples_dataset
        # create adata object to run DE on the approx counts
        # X here will be overwritten
        adata_approx = AnnData(X=adata_de.X, obs=adata_de.obs, var=adata_de.var)
        de_keys = {}
        models = [model for model in pp_samples.data_vars if model != DATA_VAR_RAW]
        for model in models:
            if model not in de_keys:
                de_keys[model] = []
            for k in range(n_samples):
                one_sample = pp_samples[model].isel(samples=k)
                # overwrite X with the posterior predictive sample
                # This allows us to save all the DE results in the same adata object
                one_sample_data = (
                    one_sample.data.to_scipy_sparse().tocsr()
                    if isinstance(one_sample.data, SparseArray)
                    else one_sample
                )
                adata_approx.X = one_sample_data.copy()
                sc.pp.normalize_total(adata_approx, target_sum=cell_scale_factor)
                sc.pp.log1p(adata_approx)

                # run DE with the imputed normalized data
                with warnings.catch_warnings():
                    warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
                    key_added = f"{UNS_NAME_RGG_PPC}_{model}_sample_{k}"
                    de_keys[model].append(key_added)
                    sc.tl.rank_genes_groups(
                        adata_approx,
                        de_groupby,
                        use_raw=False,
                        method=de_method,
                        key_added=key_added,
                    )

        groups = adata_de.obs[de_groupby].astype("category").cat.categories
        cell_counts = adata_de.obs[de_groupby].value_counts()
        df = pd.DataFrame(
            index=np.arange(len(groups) * len(models)),
            columns=[
                "gene_f1",
                "lfc_mae",
                "lfc_pearson",
                "lfc_spearman",
                "roc_auc",
                "pr_auc",
                "group",
                "model",
                "n_cells",
            ],
        )
        # Initialize storage for metrics
        self.metrics[METRIC_DIFF_EXP] = {"lfc_per_model_per_group": {}}

        for i, group in enumerate(groups):
            raw_group_data = sc.get.rank_genes_groups_df(
                adata_de, group=group, key=UNS_NAME_RGG_RAW
            )
            raw_group_data.set_index("names", inplace=True)

            for model, model_keys in de_keys.items():
                # Storage for metrics across samples
                gene_overlap_f1s, lfc_maes, lfc_pearsons, lfc_spearmans = [], [], [], []
                roc_aucs, pr_aucs, rgds, sgds = [], [], [], []
                for de_key in model_keys:
                    sample_group_data = sc.get.rank_genes_groups_df(
                        adata_approx, group=group, key=de_key
                    )
                    sample_group_data.set_index("names", inplace=True)

                    # Gene Overlap F1
                    all_genes = raw_group_data.index
                    top_genes_raw = raw_group_data[:n_top_genes_fallback].index
                    top_genes_sample = sample_group_data[:n_top_genes_fallback].index
                    true_genes = np.isin(all_genes, top_genes_raw).astype(int)
                    pred_genes = np.isin(all_genes, top_genes_sample).astype(int)
                    gene_overlap_f1s.append(
                        precision_recall_fscore_support(true_genes, pred_genes, average="binary")[
                            2
                        ]
                    )
                    # Log-fold change (LFC) metrics
                    sample_group_data = sample_group_data.reindex(raw_group_data.index)
                    rgd, sgd = (
                        raw_group_data["logfoldchanges"],
                        sample_group_data["logfoldchanges"],
                    )
                    rgds.append(rgd)
                    sgds.append(sgd)
                    lfc_maes.append(np.mean(np.abs(rgd - sgd)))
                    lfc_pearsons.append(pearsonr(rgd, sgd)[0])
                    lfc_spearmans.append(spearmanr(rgd, sgd)[0])

                    # ROC and PR metrics
                    raw_adj_p_vals = raw_group_data["pvals_adj"]
                    true = raw_adj_p_vals < p_val_thresh
                    pred = sample_group_data["scores"]

                    # Fallback for no true DE genes and most genes DE.
                    if true.sum() == 0 or true.sum() > (0.5 * len(true)):
                        true = np.zeros_like(pred)
                        true[np.argsort(raw_adj_p_vals)[:n_top_genes_fallback]] = 1

                    roc_aucs.append(roc_auc_score(true, pred))
                    pr_aucs.append(average_precision_score(true, pred))

                # Compute means over samples
                df.loc[i, "model"] = model
                df.loc[i, "group"] = group
                df.loc[i, "gene_f1"] = np.mean(gene_overlap_f1s)
                df.loc[i, "lfc_mae"] = np.mean(lfc_maes)
                df.loc[i, "lfc_pearson"] = np.mean(lfc_pearsons)
                df.loc[i, "lfc_spearman"] = np.mean(lfc_spearmans)
                df.loc[i, "roc_auc"] = np.mean(roc_aucs)
                df.loc[i, "pr_auc"] = np.mean(pr_aucs)
                df.loc[i, "n_cells"] = cell_counts[group]

                # Store LFCs for raw vs approx
                rgd_avg, sgd_avg = pd.DataFrame(rgds).mean(axis=0), pd.DataFrame(sgds).mean(axis=0)
                self.metrics[METRIC_DIFF_EXP]["lfc_per_model_per_group"].setdefault(model, {})
                self.metrics[METRIC_DIFF_EXP]["lfc_per_model_per_group"][model][str(group)] = (
                    pd.DataFrame([rgd_avg, sgd_avg], index=["raw", "approx"]).T
                )

        self.metrics[METRIC_DIFF_EXP]["summary"] = df



import os
import subprocess

import pytest
import torch


@pytest.mark.multigpu
def test_scvi_train_ddp(save_path: str):
    training_code = """
import torch
import scvi
from scvi.model import SCVI

adata = scvi.data.synthetic_iid()
SCVI.setup_anndata(adata)

model = SCVI(adata)

model.train(
    max_epochs=1,
    check_val_every_n_epoch=1,
    accelerator="gpu",
    devices=-1,
    strategy="ddp_find_unused_parameters_true",
)

torch.distributed.destroy_process_group()

assert model.is_trained
"""
    # Define the file path for the temporary script in the current working directory
    temp_file_path = os.path.join(save_path, "train_scvi_ddp_temp.py")

    # Write the training code to the file in the current working directory
    with open(temp_file_path, "w") as temp_file:
        temp_file.write(training_code)
        print(f"Temporary Python file created at: {temp_file_path}")

    def launch_ddp(world_size, temp_file_path):
        # Command to run the script via torchrun
        command = [
            "torchrun",
            "--nproc_per_node=" + str(world_size),  # Specify the number of GPUs
            temp_file_path,  # Your original script
        ]
        # Use subprocess to run the command
        try:
            # Run the command, wait for it to finish & clean up the temporary file
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            os.remove(temp_file_path)
            print(f"Error occurred while running the DDP training: {e}")
            raise
        finally:
            os.remove(temp_file_path)

    launch_ddp(torch.cuda.device_count(), temp_file_path)


@pytest.mark.multigpu
def test_scanvi_train_ddp(save_path: str):
    training_code = """
import torch
import scvi
from scvi.model import SCANVI

adata = scvi.data.synthetic_iid()
SCANVI.setup_anndata(
    adata,
    "labels",
    "label_0",
    batch_key="batch",
)

model = SCANVI(adata, n_latent=10)

model.train(
    max_epochs=100,
    train_size=0.5,
    check_val_every_n_epoch=1,
    accelerator="gpu",
    devices=-1,
    strategy="ddp_find_unused_parameters_true",
)

torch.distributed.destroy_process_group()

assert model.is_trained
"""
    # Define the file path for the temporary script in the current working directory
    temp_file_path = os.path.join(save_path, "train_scanvi_ddp_temp.py")

    # Write the training code to the file in the current working directory
    with open(temp_file_path, "w") as temp_file:
        temp_file.write(training_code)
        print(f"Temporary Python file created at: {temp_file_path}")

    def launch_ddp(world_size, temp_file_path):
        # Command to run the script via torchrun
        command = [
            "torchrun",
            "--nproc_per_node=" + str(world_size),  # Specify the number of GPUs
            temp_file_path,  # Your original script
        ]
        # Use subprocess to run the command
        try:
            # Run the command, wait for it to finish & clean up the temporary file
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            os.remove(temp_file_path)
            print(f"Error occurred while running the DDP training: {e}")
            raise
        finally:
            os.remove(temp_file_path)

    launch_ddp(torch.cuda.device_count(), temp_file_path)



import torch

from scvi import REGISTRY_KEYS
from scvi.module import Classifier
from scvi.train import AdversarialTrainingPlan


class GIMVITrainingPlan(AdversarialTrainingPlan):
    """gimVI training plan."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if kwargs["adversarial_classifier"] is True:
            self.n_output_classifier = 2
            self.adversarial_classifier = Classifier(
                n_input=self.module.n_latent,
                n_hidden=32,
                n_labels=self.n_output_classifier,
                n_layers=3,
                logits=True,
            )
        else:
            self.adversarial_classifier = kwargs["adversarial_classifier"]
        self.automatic_optimization = False
        self.validation_step_outputs = []

    def training_step(self, batch, batch_idx):
        """Training step."""
        kappa = (
            1 - self.kl_weight
            if self.scale_adversarial_loss == "auto"
            else self.scale_adversarial_loss
        )
        opts = self.optimizers()
        if not isinstance(opts, list):
            opt1 = opts
            opt2 = None
        else:
            opt1, opt2 = opts
        # batch contains both data loader outputs
        loss_output_objs = []
        n_obs = 0
        zs = []
        for i, tensors in enumerate(batch):
            n_obs += tensors[REGISTRY_KEYS.X_KEY].shape[0]
            self.loss_kwargs.update({"kl_weight": self.kl_weight, "mode": i})
            inference_kwargs = {"mode": i}
            generative_kwargs = {"mode": i}
            inference_outputs, _, loss_output = self.forward(
                tensors,
                loss_kwargs=self.loss_kwargs,
                inference_kwargs=inference_kwargs,
                generative_kwargs=generative_kwargs,
            )
            zs.append(inference_outputs["z"])
            loss_output_objs.append(loss_output)

        loss = sum([scl.loss for scl in loss_output_objs])
        loss /= n_obs
        rec_loss = sum([scl.reconstruction_loss_sum for scl in loss_output_objs])
        kl = sum([scl.kl_local_sum for scl in loss_output_objs])

        # fool classifier if doing adversarial training
        batch_tensor = [
            torch.zeros((z.shape[0], 1), device=z.device) + i for i, z in enumerate(zs)
        ]
        if kappa > 0 and self.adversarial_classifier is not False:
            fool_loss = self.loss_adversarial_classifier(
                torch.cat(zs), torch.cat(batch_tensor).long(), False
            )
            loss += fool_loss * kappa
        opt1.zero_grad()
        self.manual_backward(loss)
        opt1.step()
        return_dict = {
            "loss": loss,
            "reconstruction_loss_sum": rec_loss,
            "kl_local_sum": kl,
            "kl_global": 0.0,
            "n_obs": n_obs,
        }

        # train adversarial classifier
        # this condition will not be met unless self.adversarial_classifier is not False
        if opt2 is not None:
            zs = []
            for i, tensors in enumerate(batch):
                inference_inputs = self.module._get_inference_input(tensors)
                inference_inputs.update({"mode": i})
                outputs = self.module.inference(**inference_inputs)
                zs.append(outputs["z"])

            batch_tensor = [
                torch.zeros((z.shape[0], 1), device=z.device) + i for i, z in enumerate(zs)
            ]
            loss = self.loss_adversarial_classifier(
                torch.cat(zs).detach(), torch.cat(batch_tensor).long(), True
            )
            loss *= kappa
            opt2.zero_grad()
            self.manual_backward(loss)
            opt2.step()

        return return_dict

    def validation_step(self, batch, batch_idx, dataloader_idx):
        """Validation step."""
        self.loss_kwargs.update({"kl_weight": self.kl_weight, "mode": dataloader_idx})
        inference_kwargs = {"mode": dataloader_idx}
        generative_kwargs = {"mode": dataloader_idx}
        _, _, loss_output = self.forward(
            batch,
            loss_kwargs=self.loss_kwargs,
            inference_kwargs=inference_kwargs,
            generative_kwargs=generative_kwargs,
        )
        reconstruction_loss = loss_output.reconstruction_loss_sum
        self.validation_step_outputs.append(
            {
                "reconstruction_loss_sum": reconstruction_loss,
                "kl_local_sum": loss_output.kl_local_sum,
                "kl_global": loss_output.kl_global,
                "n_obs": loss_output.n_obs_minibatch,
            }
        )

    def on_validation_epoch_end(self):
        """Aggregate validation step information."""
        super().on_validation_epoch_end()
        outputs = self.validation_step_outputs
        n_obs, elbo, rec_loss, kl_local = 0, 0, 0, 0
        for val_metrics in outputs:
            elbo += val_metrics["reconstruction_loss_sum"] + val_metrics["kl_local_sum"]
            rec_loss += val_metrics["reconstruction_loss_sum"]
            kl_local += val_metrics["kl_local_sum"]
            n_obs += val_metrics["n_obs"]
        # kl global same for each minibatch
        self.log("elbo_validation", elbo / n_obs)
        self.log("reconstruction_loss_validation", rec_loss / n_obs)
        self.log("kl_local_validation", kl_local / n_obs)
        self.log("kl_global_validation", 0.0)
        self.validation_step_outputs.clear()  # free memory



from ray import tune
from ray.tune import ResultGrid

from scvi import settings
from scvi.autotune import AutotuneExperiment, run_autotune
from scvi.data import synthetic_iid
from scvi.model import SCVI


def test_run_autotune_scvi_basic(save_path: str):
    settings.logging_dir = save_path
    adata = synthetic_iid()
    SCVI.setup_anndata(adata)

    experiment = run_autotune(
        SCVI,
        adata,
        metrics=["elbo_validation"],
        mode="min",
        search_space={
            "model_params": {
                "n_hidden": tune.choice([1, 2]),
            },
            "train_params": {
                "max_epochs": 1,
            },
        },
        num_samples=1,
        seed=0,
        scheduler="asha",
        searcher="hyperopt",
    )
    assert isinstance(experiment, AutotuneExperiment)
    assert hasattr(experiment, "result_grid")
    assert isinstance(experiment.result_grid, ResultGrid)


def test_run_autotune_scvi_no_anndata(save_path: str, n_batches: int = 3):
    from scvi.dataloaders import DataSplitter

    settings.logging_dir = save_path
    adata = synthetic_iid(n_batches=n_batches)
    SCVI.setup_anndata(adata, batch_key="batch")
    manager = SCVI._get_most_recent_anndata_manager(adata)

    datamodule = DataSplitter(manager)
    datamodule.n_vars = adata.n_vars
    datamodule.n_batch = n_batches

    experiment = run_autotune(
        SCVI,
        datamodule,
        metrics=["elbo_validation"],
        mode="min",
        search_space={
            "model_params": {
                "n_hidden": tune.choice([1, 2]),
            },
            "train_params": {
                "max_epochs": 1,
            },
        },
        num_samples=1,
        seed=0,
        scheduler="asha",
        searcher="hyperopt",
    )
    assert isinstance(experiment, AutotuneExperiment)
    assert hasattr(experiment, "result_grid")
    assert isinstance(experiment.result_grid, ResultGrid)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import torch

from scvi import REGISTRY_KEYS
from scvi.data import AnnDataManager
from scvi.data.fields import CategoricalObsField, LayerField, NumericalObsField
from scvi.external.stereoscope._module import RNADeconv, SpatialDeconv
from scvi.model.base import BaseModelClass, UnsupervisedTrainingMixin
from scvi.utils import setup_anndata_dsp
from scvi.utils._docstrings import devices_dsp

if TYPE_CHECKING:
    from typing import Literal

    from anndata import AnnData

logger = logging.getLogger(__name__)


class RNAStereoscope(UnsupervisedTrainingMixin, BaseModelClass):
    """Reimplementation of Stereoscope :cite:p:`Andersson20`.

    Deconvolution of spatial transcriptomics from single-cell transcriptomics. Original
    implementation: https://github.com/almaan/stereoscope.

    Parameters
    ----------
    sc_adata
        single-cell AnnData object that has been registered via
        :meth:`~scvi.external.RNAStereoscope.setup_anndata`.
    **model_kwargs
        Keyword args for :class:`~scvi.external.stereoscope.RNADeconv`

    Examples
    --------
    >>> sc_adata = anndata.read_h5ad(path_to_sc_anndata)
    >>> scvi.external.RNAStereoscope.setup_anndata(sc_adata, labels_key="labels")
    >>> stereo = scvi.external.stereoscope.RNAStereoscope(sc_adata)
    >>> stereo.train()

    Notes
    -----
    See further usage examples in the following tutorial:

    1. :doc:`/tutorials/notebooks/spatial/stereoscope_heart_LV_tutorial`
    """

    def __init__(
        self,
        sc_adata: AnnData,
        **model_kwargs,
    ):
        super().__init__(sc_adata)
        self.n_genes = self.summary_stats.n_vars
        self.n_labels = self.summary_stats.n_labels
        # first we have the scRNA-seq model
        self.module = RNADeconv(
            n_genes=self.n_genes,
            n_labels=self.n_labels,
            **model_kwargs,
        )
        self._model_summary_string = (
            f"RNADeconv Model with params: \nn_genes: {self.n_genes}, n_labels: {self.n_labels}"
        )
        self.init_params_ = self._get_init_params(locals())

    @devices_dsp.dedent
    def train(
        self,
        max_epochs: int = 400,
        lr: float = 0.01,
        accelerator: str = "auto",
        devices: int | list[int] | str = "auto",
        train_size: float = 1,
        validation_size: float | None = None,
        shuffle_set_split: bool = True,
        batch_size: int = 128,
        datasplitter_kwargs: dict | None = None,
        plan_kwargs: dict | None = None,
        **kwargs,
    ):
        """Trains the model using MAP inference.

        Parameters
        ----------
        max_epochs
            Number of epochs to train for
        lr
            Learning rate for optimization.
        %(param_accelerator)s
        %(param_devices)s
        train_size
            Size of training set in the range [0.0, 1.0].
        validation_size
            Size of the test set. If `None`, defaults to 1 - `train_size`. If
            `train_size + validation_size < 1`, the remaining cells belong to a test set.
        shuffle_set_split
            Whether to shuffle indices before splitting. If `False`, the val, train, and test set
            are split in the sequential order of the data according to `validation_size` and
            `train_size` percentages.
        batch_size
            Minibatch size to use during training.
        datasplitter_kwargs
            Additional keyword arguments passed into :class:`~scvi.dataloaders.DataSplitter`.
        plan_kwargs
            Keyword args for :class:`~scvi.train.TrainingPlan`. Keyword arguments passed to
            `train()` will overwrite values present in `plan_kwargs`, when appropriate.
        **kwargs
            Other keyword args for :class:`~scvi.train.Trainer`.
        """
        update_dict = {
            "lr": lr,
        }
        if plan_kwargs is not None:
            plan_kwargs.update(update_dict)
        else:
            plan_kwargs = update_dict

        super().train(
            max_epochs=max_epochs,
            accelerator=accelerator,
            devices=devices,
            train_size=train_size,
            validation_size=validation_size,
            shuffle_set_split=shuffle_set_split,
            batch_size=batch_size,
            datasplitter_kwargs=datasplitter_kwargs,
            plan_kwargs=plan_kwargs,
            **kwargs,
        )

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        labels_key: str | None = None,
        layer: str | None = None,
        **kwargs,
    ):
        """%(summary)s.

        Parameters
        ----------
        %(param_labels_key)s
        %(param_layer)s
        """
        setup_method_args = cls._get_setup_method_args(**locals())
        anndata_fields = [
            LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            CategoricalObsField(REGISTRY_KEYS.LABELS_KEY, labels_key),
        ]
        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)


class SpatialStereoscope(UnsupervisedTrainingMixin, BaseModelClass):
    """Reimplementation of Stereoscope :cite:p:`Andersson20`.

    Deconvolution of spatial transcriptomics from single-cell transcriptomics. Original
    implementation: https://github.com/almaan/stereoscope.

    Parameters
    ----------
    st_adata
        spatial transcriptomics AnnData object that has been registered via
        :meth:`~scvi.external.SpatialStereoscope.setup_anndata`.
    sc_params
        parameters of the model learned from the single-cell RNA seq data for deconvolution.
    cell_type_mapping
        numpy array mapping for the cell types used in the deconvolution
    prior_weight
        how to reweight the minibatches for stochastic optimization. "n_obs" is the valid
        procedure, "minibatch" is the procedure implemented in Stereoscope.
    **model_kwargs
        Keyword args for :class:`~scvi.external.stereoscope.SpatialDeconv`

    Examples
    --------
    >>> sc_adata = anndata.read_h5ad(path_to_sc_anndata)
    >>> scvi.external.RNAStereoscope.setup_anndata(sc_adata, labels_key="labels")
    >>> sc_model = scvi.external.stereoscope.RNAStereoscope(sc_adata)
    >>> sc_model.train()
    >>> st_adata = anndata.read_h5ad(path_to_st_anndata)
    >>> scvi.external.SpatialStereoscope.setup_anndata(st_adata)
    >>> stereo = scvi.external.stereoscope.SpatialStereoscope.from_rna_model(st_adata, sc_model)
    >>> stereo.train()
    >>> st_adata.obsm["deconv"] = stereo.get_proportions()

    Notes
    -----
    See further usage examples in the following tutorials:

    1. :doc:`/tutorials/notebooks/spatial/stereoscope_heart_LV_tutorial`
    """

    def __init__(
        self,
        st_adata: AnnData,
        sc_params: tuple[np.ndarray],
        cell_type_mapping: np.ndarray,
        prior_weight: Literal["n_obs", "minibatch"] = "n_obs",
        **model_kwargs,
    ):
        super().__init__(st_adata)
        self.module = SpatialDeconv(
            n_spots=st_adata.n_obs,
            sc_params=sc_params,
            prior_weight=prior_weight,
            **model_kwargs,
        )
        self._model_summary_string = f"RNADeconv Model with params: \nn_spots: {st_adata.n_obs}"
        self.cell_type_mapping = cell_type_mapping
        self.init_params_ = self._get_init_params(locals())

    @classmethod
    def from_rna_model(
        cls,
        st_adata: AnnData,
        sc_model: RNAStereoscope,
        prior_weight: Literal["n_obs", "minibatch"] = "n_obs",
        **model_kwargs,
    ):
        """Alternate constructor for exploiting a pre-trained model on RNA-seq data.

        Parameters
        ----------
        st_adata
            registed anndata object
        sc_model
            trained RNADeconv model
        prior_weight
            how to reweight the minibatches for stochastic optimization. "n_obs" is the valid
            procedure, "minibatch" is the procedure implemented in Stereoscope.
        **model_kwargs
            Keyword args for :class:`~scvi.external.SpatialDeconv`
        """
        return cls(
            st_adata,
            sc_model.module.get_params(),
            sc_model.adata_manager.get_state_registry(
                REGISTRY_KEYS.LABELS_KEY
            ).categorical_mapping,
            prior_weight=prior_weight,
            **model_kwargs,
        )

    def get_proportions(self, keep_noise=False) -> pd.DataFrame:
        """Returns the estimated cell type proportion for the spatial data.

        Shape is n_cells x n_labels OR n_cells x (n_labels + 1) if keep_noise

        Parameters
        ----------
        keep_noise
            whether to account for the noise term as a standalone cell type in the proportion
            estimate.
        """
        self._check_if_trained()

        column_names = self.cell_type_mapping
        if keep_noise:
            column_names = column_names.append("noise_term")
        return pd.DataFrame(
            data=self.module.get_proportions(keep_noise),
            columns=column_names,
            index=self.adata.obs.index,
        )

    def get_scale_for_ct(
        self,
        y: np.ndarray,
    ) -> np.ndarray:
        r"""Calculate the cell type specific expression.

        Parameters
        ----------
        y
            numpy array containing the list of cell types

        Returns
        -------
        gene_expression
        """
        self._check_if_trained()
        ind_y = np.array([np.where(ct == self.cell_type_mapping)[0][0] for ct in y])
        if ind_y.shape != y.shape:
            raise ValueError(
                "Incorrect shape after matching cell types to reference mapping. Please check "
                "cell type query."
            )
        px_scale = self.module.get_ct_specific_expression(torch.tensor(ind_y)[:, None])
        return np.array(px_scale.cpu())

    @devices_dsp.dedent
    def train(
        self,
        max_epochs: int = 400,
        lr: float = 0.01,
        accelerator: str = "auto",
        devices: int | list[int] | str = "auto",
        shuffle_set_split: bool = True,
        batch_size: int = 128,
        datasplitter_kwargs: dict | None = None,
        plan_kwargs: dict | None = None,
        **kwargs,
    ):
        """Trains the model using MAP inference.

        Parameters
        ----------
        max_epochs
            Number of epochs to train for
        lr
            Learning rate for optimization.
        %(param_accelerator)s
        %(param_devices)s
        shuffle_set_split
            Whether to shuffle indices before splitting. If `False`, the val, train, and test set
            are split in the sequential order of the data according to `validation_size` and
            `train_size` percentages.
        batch_size
            Minibatch size to use during training.
        datasplitter_kwargs
            Additional keyword arguments passed into :class:`~scvi.dataloaders.DataSplitter`.
        plan_kwargs
            Keyword args for :class:`~scvi.train.TrainingPlan`. Keyword arguments passed to
            `train()` will overwrite values present in `plan_kwargs`, when appropriate.
        **kwargs
            Other keyword args for :class:`~scvi.train.Trainer`.
        """
        update_dict = {
            "lr": lr,
        }
        if plan_kwargs is not None:
            plan_kwargs.update(update_dict)
        else:
            plan_kwargs = update_dict
        super().train(
            max_epochs=max_epochs,
            accelerator=accelerator,
            devices=devices,
            train_size=1,
            validation_size=None,
            shuffle_set_split=shuffle_set_split,
            batch_size=batch_size,
            datasplitter_kwargs=datasplitter_kwargs,
            plan_kwargs=plan_kwargs,
            **kwargs,
        )

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        layer: str | None = None,
        **kwargs,
    ):
        """%(summary)s.

        Parameters
        ----------
        %(param_layer)s
        """
        setup_method_args = cls._get_setup_method_args(**locals())
        # add index for each cell (provided to pyro plate for correct minibatching)
        adata.obs["_indices"] = np.arange(adata.n_obs)
        anndata_fields = [
            LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            NumericalObsField(REGISTRY_KEYS.INDICES_KEY, "_indices"),
        ]
        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)



from anndata import read_h5ad
from anndata.io import read_csv, read_loom, read_text

from ._anntorchdataset import AnnTorchDataset
from ._datasets import (
    annotation_simulation,
    brainlarge_dataset,
    breast_cancer_dataset,
    cellxgene,
    cortex,
    dataset_10x,
    frontalcortex_dropseq,
    heart_cell_atlas_subsampled,
    mouse_ob_dataset,
    pbmc_dataset,
    pbmc_seurat_v4_cite_seq,
    pbmcs_10x_cite_seq,
    prefrontalcortex_starmap,
    purified_pbmc_dataset,
    retina,
    smfish,
    spleen_lymph_cite_seq,
    synthetic_iid,
)
from ._manager import AnnDataManager, AnnDataManagerValidationCheck
from ._preprocessing import (
    add_dna_sequence,
    organize_cite_seq_10x,
    organize_multiome_anndatas,
    poisson_gene_selection,
    reads_to_fragments,
)
from ._read import read_10x_atac, read_10x_multiome

__all__ = [
    "AnnTorchDataset",
    "AnnDataManagerValidationCheck",
    "AnnDataManager",
    "poisson_gene_selection",
    "organize_cite_seq_10x",
    "pbmcs_10x_cite_seq",
    "spleen_lymph_cite_seq",
    "dataset_10x",
    "purified_pbmc_dataset",
    "brainlarge_dataset",
    "synthetic_iid",
    "pbmc_dataset",
    "cortex",
    "smfish",
    "breast_cancer_dataset",
    "mouse_ob_dataset",
    "retina",
    "prefrontalcortex_starmap",
    "frontalcortex_dropseq",
    "annotation_simulation",
    "read_h5ad",
    "read_csv",
    "read_loom",
    "read_text",
    "read_10x_atac",
    "read_10x_multiome",
    "heart_cell_atlas_subsampled",
    "organize_multiome_anndatas",
    "pbmc_seurat_v4_cite_seq",
    "add_dna_sequence",
    "reads_to_fragments",
    "cellxgene",
]



scvi_pretext = """
ScVI is a variational inference model for single-cell RNA-seq data that can learn an underlying
latent space, integrate technical batches and impute dropouts.
The learned low-dimensional latent representation of the data can be used for visualization and
clustering.

scVI takes as input a scRNA-seq gene expression matrix with cells and genes.
We provide an extensive [user guide](https://docs.scvi-tools.org/en/1.2.0/user_guide/models/scvi.html).

- See our original manuscript for further details of the model:
[scVI manuscript](https://www.nature.com/articles/s41592-018-0229-2).
- See our manuscript on [scvi-hub](https://www.biorxiv.org/content/10.1101/2024.03.01.582887v2) how
to leverage pre-trained models.

This model can be used for fine tuning on new data using our Arches framework:
[Arches tutorial](https://docs.scvi-tools.org/en/1.0.0/tutorials/notebooks/scarches_scvi_tools.html).
"""

scanvi_pretext = """
ScANVI is a variational inference model for single-cell RNA-seq data that can learn an underlying
latent space, integrate technical batches and impute dropouts.
In addition, to scVI, ScANVI is a semi-supervised model that can leverage labeled data to learn a
cell-type classifier in the latent space and afterward predict cell types of new data.
The learned low-dimensional latent representation of the data can be used for visualization and
clustering.

scANVI takes as input a scRNA-seq gene expression matrix with cells and genes as well as a
cell-type annotation for a subset of cells.
We provide an extensive [user guide](https://docs.scvi-tools.org/en/1.2.0/user_guide/models/scanvi.html).

- See our original manuscript for further details of the model:
[scANVI manuscript](https://www.embopress.org/doi/full/10.15252/msb.20209620).
- See our manuscript on [scvi-hub](https://www.biorxiv.org/content/10.1101/2024.03.01.582887v2)
how to leverage pre-trained models.

This model can be used for fine tuning on new data using our Arches framework:
[Arches tutorial](https://docs.scvi-tools.org/en/1.0.0/tutorials/notebooks/scarches_scvi_tools.html).
"""

condscvi_pretext = """
CondSCVI is a variational inference model for single-cell RNA-seq data that can learn an underlying
latent space. The predictions of the model are meant to be afterward
used for deconvolution of a second spatial transcriptomics dataset in DestVI. DestVI predicts the
cell-type proportions as well as cell type-specific activation state
in the spatial data.

CondSCVI takes as input a scRNA-seq gene expression matrix with cells and genes as well as a
cell-type annotation for all cells.
We provide an extensive [user guide](https://docs.scvi-tools.org/en/1.2.0/user_guide/models/destvi.html)
for DestVI including a description of CondSCVI.

- See our original manuscript for further details of the model:
[DestVI manuscript](https://www.nature.com/articles/s41587-022-01272-8).
- See our manuscript on [scvi-hub](https://www.biorxiv.org/content/10.1101/2024.03.01.582887v2)
how to leverage pre-trained models.
"""

stereoscope_pretext = """
Stereoscope is a variational inference model for single-cell RNA-seq data that can learn a
cell-type specific rate of gene expression. The predictions of the model are meant to be afterward
used for deconvolution of a second spatial transcriptomics dataset in Stereoscope. Stereoscope
predicts the cell-type proportions in the spatial data.

Stereoscope takes as input a scRNA-seq gene expression matrix with cells and genes as well as a
cell-type annotation for all cells.
We provide an extensive for DestVI including a description of CondSCVI
[user guide](https://docs.scvi-tools.org/en/1.2.0/user_guide/models/destvi.html).

- See our original manuscript for further details of the model:
[Stereoscope manuscript](https://www.nature.com/articles/s42003-020-01247-y) as well as the
[scvi-tools manuscript](https://www.nature.com/articles/s41587-021-01206-w) about implementation
details.
- See our manuscript on [scvi-hub](https://www.biorxiv.org/content/10.1101/2024.03.01.582887v2)
how to leverage pre-trained models.
"""

totalvi_pretext = """
TotalVI is a variational inference model for single-cell RNA-seq as well as protein data that can
learn an underlying latent space, integrate technical batches, impute dropouts,
and predict protein expression given gene expression or missing protein data given gene expression
and protein data for a subset of proteins.
The learned low-dimensional latent representation of the data can be used for visualization and
clustering.

TotalVI takes as input a scRNA-seq gene expression and protein expression matrix with cells and
genes.
We provide an extensive [user guide](https://docs.scvi-tools.org/en/1.2.0/user_guide/models/totalvi.html).

- See our original manuscript for further details of the model:
[TotalVI manuscript](https://www.nature.com/articles/s41592-020-01050-x).
- See our manuscript on [scvi-hub](https://www.biorxiv.org/content/10.1101/2024.03.01.582887v2)
how to leverage pre-trained models.

This model can be used for fine tuning on new data using our Arches framework:
[Arches tutorial](https://docs.scvi-tools.org/en/1.0.0/tutorials/notebooks/scarches_scvi_tools.html).
"""


template = """\
---
{card_data}
---

{pretext}

# Model Description

{description}

# Metrics

We provide here key performance metrics for the uploaded model, if provided by the data uploader.

<details>
<summary><strong>Coefficient of variation</strong></summary>

The cell-wise coefficient of variation summarizes how well variation between different cells is
preserved by the generated model expression. Below a squared Pearson correlation coefficient of 0.4
, we would recommend not to use generated data for downstream analysis, while the generated latent
space might still be useful for analysis.

**Cell-wise Coefficient of Variation**:

{cell_wise_cv}

The gene-wise coefficient of variation summarizes how well variation between different genes is
preserved by the generated model expression. This value is usually quite high.

**Gene-wise Coefficient of Variation**:

{gene_wise_cv}

</details>

<details>
<summary><strong>Differential expression metric</strong></summary>

The differential expression metric provides a summary of the differential expression analysis
between cell types or input clusters. We provide here the F1-score, Pearson Correlation
Coefficient of Log-Foldchanges, Spearman Correlation Coefficient, and Area Under the Precision
Recall Curve (AUPRC) for the differential expression analysis using Wilcoxon Rank Sum test for each
cell-type.

**Differential expression**:

{de_metrics}

</details>

# Model Properties

We provide here key parameters used to setup and train the model.

<details>
<summary><strong>Model Parameters</strong></summary>

These provide the settings to setup the original model:
```json
{model_init_params}
```

</details>

<details>
<summary><strong>Setup Data Arguments</strong></summary>

Arguments passed to setup_anndata of the original model:
```json
{model_setup_anndata_args}
```

</details>

<details>
<summary><strong>Data Registry</strong></summary>

Registry elements for AnnData manager:
{model_data_registry}

- **Data is Minified**: {data_is_minified}

</details>

<details>
<summary><strong>Summary Statistics</strong></summary>

{model_summary_stats}

</details>


<details>
<summary><strong>Training</strong></summary>

<!-- If your model is not uploaded with any data (e.g., minified data) on the Model Hub, then make
sure to provide this field if you want users to be able to access your training data. See the
scvi-tools documentation for details. -->
**Training data url**: {training_data_url}

If provided by the original uploader, for those interested in understanding or replicating the
training process, the code is available at the link below.

**Training Code URL**: {training_code_url}

</details>
\n
# References

{references}
"""



import numpy as np
import pytest

from scvi.data import synthetic_iid


@pytest.mark.parametrize("sparse_format", ["csr_matrix", "csc_matrix", None])
def test_synthetic_iid_sparse_format(sparse_format: str | None):
    _ = synthetic_iid(sparse_format=sparse_format)


@pytest.mark.parametrize("return_mudata", [False, True])
def test_synthetic_iid_coords(return_mudata: bool):
    adata = synthetic_iid(return_mudata=return_mudata, generate_coordinates=True)
    assert "coordinates" in adata.obsm
    assert isinstance(adata.obsm["coordinates"], np.ndarray)
    assert adata.obsm["coordinates"].shape == (adata.n_obs, 2)



import mudata
import numpy as np
import pytest

from scvi.data import synthetic_iid
from scvi.external import Tangram

modalities = {"density_prior_key": "sp", "sc_layer": "sc", "sp_layer": "sp"}


def _get_mdata(sparse_format: str | None = None):
    dataset1 = synthetic_iid(batch_size=100, sparse_format=sparse_format)
    dataset2 = dataset1[-25:].copy()
    dataset1 = dataset1[:-25].copy()
    mdata = mudata.MuData({"sc": dataset1, "sp": dataset2})
    ad_sp = mdata.mod["sp"]
    rna_count_per_spot = np.asarray(ad_sp.X.sum(axis=1)).squeeze()
    ad_sp.obs["rna_count_based_density"] = rna_count_per_spot / np.sum(rna_count_per_spot)
    ad_sp.obs["bad_prior"] = np.random.uniform(size=ad_sp.n_obs)
    return mdata


@pytest.mark.parametrize(
    ("density_prior_key", "constrained"),
    [
        (None, False),
        ("rna_count_based_density", False),
        ("rna_count_based_density", True),
    ],
)
def test_tangram(density_prior_key, constrained):
    mdata = _get_mdata()
    Tangram.setup_mudata(
        mdata,
        density_prior_key=density_prior_key,
        modalities=modalities,
    )
    if constrained:
        target_count = 2
    else:
        target_count = None
    model = Tangram(mdata, constrained=constrained, target_count=target_count)
    model.train(max_epochs=1)
    mdata.mod["sc"].obsm["mapper"] = model.get_mapper_matrix()
    model.project_cell_annotations(
        mdata.mod["sc"],
        mdata.mod["sp"],
        mdata.mod["sc"].obsm["mapper"],
        mdata.mod["sc"].obs.labels,
    )
    model.project_genes(mdata.mod["sc"], mdata.mod["sp"], mdata.mod["sc"].obsm["mapper"])


def test_tangram_errors():
    mdata = _get_mdata()
    Tangram.setup_mudata(
        mdata,
        density_prior_key="rna_count_based_density",
        modalities=modalities,
    )
    with pytest.raises(ValueError):
        Tangram(mdata, constrained=True, target_count=None)

    Tangram.setup_mudata(
        mdata,
        density_prior_key="bad_prior",
        modalities=modalities,
    )
    with pytest.raises(ValueError):
        Tangram(mdata)



import numpy as np
import pytest
from mudata import MuData

import scvi
from scvi.data import synthetic_iid
from scvi.data._constants import ADATA_MINIFY_TYPE
from scvi.data._utils import _is_minified
from scvi.model import TOTALVI

OBSERVED_LIB_SIZE = "observed_lib_size"


def prep_model(cls=TOTALVI, use_size_factor=False):
    # create a synthetic dataset
    adata = synthetic_iid()
    adata_counts = adata.X
    if use_size_factor:
        adata.obs["size_factor"] = np.random.randint(1, 5, size=(adata.shape[0],))
    adata.var["n_counts"] = np.squeeze(np.asarray(np.sum(adata_counts, axis=0)))
    adata.varm["my_varm"] = np.random.negative_binomial(5, 0.3, size=(adata.shape[1], 3))
    adata.layers["my_layer"] = np.ones_like(adata.X)
    adata_before_setup = adata.copy()

    # run setup_anndata
    setup_kwargs = {
        "batch_key": "batch",
        "protein_expression_obsm_key": "protein_expression",
        "protein_names_uns_key": "protein_names",
    }
    if use_size_factor:
        setup_kwargs["size_factor_key"] = "size_factor"
    cls.setup_anndata(
        adata,
        **setup_kwargs,
    )

    # create and train the model
    if cls == TOTALVI:
        model = cls(adata, n_latent=5)
    else:
        model = cls(adata, n_latent=5, n_genes=50, n_regions=50)
    model.train(1, check_val_every_n_epoch=1, train_size=0.5)

    # get the adata lib size
    adata_lib_size = np.squeeze(np.asarray(adata_counts.sum(axis=1)))
    assert (
        np.min(adata_lib_size) > 0
    )  # make sure it's not all zeros and there are no negative values

    return model, adata, adata_lib_size, adata_before_setup


def run_test_for_model_with_minified_adata(
    cls=TOTALVI,
    n_samples: int = 1,
    give_mean: bool = False,
    use_size_factor=False,
):
    model, adata, adata_lib_size, _ = prep_model(cls, use_size_factor)

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv
    adata_orig = adata.copy()

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR
    assert model.adata_manager.registry is model.registry_

    # make sure the original adata we set up the model with was not changed
    assert adata is not model.adata
    assert _is_minified(adata) is False

    assert adata_orig.layers.keys() == model.adata.layers.keys()
    orig_obs_df = adata_orig.obs
    obs_keys = OBSERVED_LIB_SIZE
    orig_obs_df[obs_keys] = adata_lib_size
    assert model.adata.obs.equals(orig_obs_df)
    assert model.adata.var_names.equals(adata_orig.var_names)
    assert model.adata.var.equals(adata_orig.var)
    assert model.adata.varm.keys() == adata_orig.varm.keys()
    np.testing.assert_array_equal(model.adata.varm["my_varm"], adata_orig.varm["my_varm"])


def prep_model_mudata(cls=TOTALVI, use_size_factor=False, layer=None):
    # create a synthetic dataset
    mdata = synthetic_iid(return_mudata=True)
    if use_size_factor:
        mdata.obs["size_factor_rna"] = mdata["rna"].X.sum(1)
        mdata.obs["size_factor_atac"] = (mdata["accessibility"].X.sum(1) + 1) / (
            np.max(mdata["accessibility"].X.sum(1)) + 1.01
        )
    if layer is not None:
        for mod in mdata.mod_names:
            mdata[mod].layers[layer] = mdata[mod].X.copy()
            mdata[mod].X = np.zeros_like(mdata[mod].X)
    mdata.var["n_counts"] = np.squeeze(
        np.concatenate(
            [
                np.asarray(np.sum(mdata["rna"].X, axis=0)),
                np.asarray(np.sum(mdata["protein_expression"].X, axis=0)),
                np.asarray(np.sum(mdata["accessibility"].X, axis=0)),
            ]
        )
    )
    mdata.varm["my_varm"] = np.random.negative_binomial(5, 0.3, size=(mdata.shape[1], 3))
    mdata_before_setup = mdata.copy()

    # run setup_anndata
    setup_kwargs = {
        "batch_key": "batch",
    }

    if use_size_factor:
        setup_kwargs["size_factor_key"] = "size_factor_rna"

    # create and train the model
    if cls == TOTALVI:
        mdata = MuData({"rna": mdata["rna"], "protein_expression": mdata["protein_expression"]})
        mdata.obs = mdata_before_setup.obs
        cls.setup_mudata(
            mdata,
            modalities={"rna_layer": "rna", "protein_layer": "protein_expression"},
            **setup_kwargs,
        )
        model = cls(mdata, n_latent=5)
    else:
        raise ValueError("Bad Model name as input to test")
    model.train(1, check_val_every_n_epoch=1, train_size=0.5)

    # get the mdata lib size
    mdata_lib_size = np.squeeze(np.asarray(mdata["rna"].X.sum(axis=1)))
    assert (
        np.min(mdata_lib_size) > 0
    )  # make sure it's not all zeros and there are no negative values

    return model, mdata, mdata_lib_size, mdata_before_setup


def run_test_for_model_with_minified_mudata(
    cls=TOTALVI,
    use_size_factor=False,
):
    model, mdata, mdata_lib_size, _ = prep_model_mudata(cls, use_size_factor)

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    mdata_orig = mdata.copy()

    model.minify_mudata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR
    assert model.adata_manager.registry is model.registry_

    # make sure the original mdata we set up the model with was not changed
    assert mdata is not model.adata
    assert _is_minified(mdata) is False
    assert _is_minified(model.adata) is True

    assert mdata_orig["rna"].layers.keys() == model.adata["rna"].layers.keys()
    orig_obs_df = mdata_orig.obs
    obs_keys = OBSERVED_LIB_SIZE
    orig_obs_df[obs_keys] = mdata_lib_size
    assert model.adata.obs.equals(orig_obs_df)
    assert model.adata.var_names.equals(mdata_orig.var_names)
    assert model.adata.var.equals(mdata_orig.var)
    assert model.adata.varm.keys() == mdata_orig.varm.keys()
    np.testing.assert_array_equal(model.adata.varm["my_varm"], mdata_orig.varm["my_varm"])


def assert_approx_equal(a, b):
    # Allclose because on GPU, the values are not exactly the same
    # as some values are moved to cpu during data minification
    np.testing.assert_allclose(a, b, rtol=3e-1, atol=5e-1)


@pytest.mark.parametrize("cls", [TOTALVI])
@pytest.mark.parametrize("use_size_factor", [True])
def test_with_minified_adata(cls, use_size_factor: bool):
    run_test_for_model_with_minified_adata(cls=cls, use_size_factor=use_size_factor)


@pytest.mark.parametrize("cls", [TOTALVI])
def test_with_minified_mdata_get_normalized_expression(cls):
    model, mdata, _, _ = prep_model_mudata(cls=cls, use_size_factor=True)

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    scvi.settings.seed = 1
    exprs_orig = model.get_normalized_expression(n_samples=500)

    model.minify_mudata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    scvi.settings.seed = 1
    exprs_new = model.get_normalized_expression(n_samples=500)

    if type(exprs_new) is tuple:
        for ii in range(len(exprs_new)):
            assert exprs_new[ii].shape == mdata[mdata.mod_names[ii]].shape
        for ii in range(len(exprs_new)):
            assert_approx_equal(exprs_new[ii], exprs_orig[ii])
    else:
        assert exprs_new.shape == exprs_orig.shape
        assert_approx_equal(exprs_new, exprs_orig)


def test_totalvi_downstream_with_minified_mdata():
    model, mdata, _, _ = prep_model_mudata(cls=TOTALVI, use_size_factor=True)
    # non-default gene list and n_samples > 1
    gl = mdata.var_names[:5].to_list()

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    scvi.settings.seed = 1

    model.minify_mudata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    scvi.settings.seed = 1
    assert model.get_normalized_expression(gene_list=gl, library_size="latent")
    assert model.get_normalized_expression(gene_list=gl, library_size=1)
    sample = model.posterior_predictive_sample()
    assert sample["rna"].shape == mdata["rna"].shape
    assert sample["protein_expression"].shape == mdata["protein_expression"].shape
    corr = model.get_feature_correlation_matrix()
    assert corr.shape == (mdata.n_vars, mdata.n_vars)
    fore = model.get_protein_foreground_probability()
    assert fore.shape == (mdata.n_obs, mdata["protein_expression"].n_vars)
    model.differential_expression(groupby="labels")


def test_totalvi_downstream_with_minified_mdata_keep_counts():
    model, mdata, _, _ = prep_model_mudata(cls=TOTALVI, use_size_factor=True)

    # non-default gene list and n_samples > 1
    gl = mdata.var_names[:5].to_list()

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    scvi.settings.seed = 1

    model.minify_mudata(minified_data_type="latent_posterior_parameters_with_counts")
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR_WITH_COUNTS

    scvi.settings.seed = 1
    assert model.get_normalized_expression(gene_list=gl, library_size="latent")
    assert model.get_normalized_expression(gene_list=gl, library_size=1)
    sample = model.posterior_predictive_sample()
    assert sample["rna"].shape == mdata["rna"].shape
    assert sample["protein_expression"].shape == mdata["protein_expression"].shape
    corr = model.get_feature_correlation_matrix()
    assert corr.shape == (mdata.n_vars, mdata.n_vars)
    fore = model.get_protein_foreground_probability()
    assert fore.shape == (mdata.n_obs, mdata["protein_expression"].n_vars)
    model.differential_expression(groupby="labels")


@pytest.mark.parametrize("cls", [TOTALVI])
def test_validate_unsupported_if_minified(cls):
    model, _, _, _ = prep_model_mudata(cls=cls, use_size_factor=True)

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    model.minify_mudata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    common_err_msg = "The {} function currently does not support minified data."

    with pytest.raises(ValueError) as e:
        model.get_elbo()
    assert str(e.value) == common_err_msg.format("VAEMixin.get_elbo")

    with pytest.raises(ValueError) as e:
        model.get_reconstruction_error()
    assert str(e.value) == common_err_msg.format("VAEMixin.get_reconstruction_error")

    with pytest.raises(ValueError) as e:
        model.get_marginal_ll()
    assert str(e.value) == common_err_msg.format("VAEMixin.get_marginal_ll")


@pytest.mark.parametrize("cls", [TOTALVI])
def test_with_minified_mdata_save_then_load(cls, save_path):
    # create a model and minify its mdata, then save it and its mdata.
    # Load it back up using the same (minified) mdata. Validate that the
    # loaded model has the minified_data_type attribute set as expected.
    model, mdata, _, _ = prep_model_mudata(cls=cls, use_size_factor=True)

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    model.minify_mudata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    model.save(save_path, overwrite=True, save_anndata=True)
    model.view_setup_args(save_path)
    # load saved model with saved (minified) mdata
    loaded_model = cls.load(save_path, adata=mdata)

    assert loaded_model.minified_data_type is None


@pytest.mark.parametrize("cls", [TOTALVI])
def test_with_minified_mdata_save_then_load_with_non_minified_mdata(cls, save_path):
    # create a model and minify its mdata, then save it and its mdata.
    # Load it back up using a non-minified mdata. Validate that the
    # loaded model does not has the minified_data_type attribute set.
    model, mdata, _, mdata_before_setup = prep_model_mudata(cls=cls, use_size_factor=True)

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    model.minify_mudata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    model.save(save_path, overwrite=True, save_anndata=False, legacy_mudata_format=True)
    # load saved model with a non-minified mdata
    loaded_model = cls.load(save_path, adata=mdata_before_setup)

    assert loaded_model.minified_data_type is None


@pytest.mark.parametrize("cls", [TOTALVI])
def test_save_then_load_with_minified_mdata(cls, save_path):
    # create a model, then save it and its mdata (non-minified).
    # Load it back up using a minified mdata. Validate that this
    # fails, as expected because we don't have a way to validate
    # whether the minified-mdata was set up correctly
    model, _, _, _ = prep_model_mudata(cls=cls, use_size_factor=True)

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    model.save(save_path, overwrite=True, save_anndata=False, legacy_mudata_format=True)

    model.minify_mudata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    # loading this model with a minified mdata is not allowed because
    # we don't have a way to validate whether the minified-mdata was
    # set up correctly
    with pytest.raises(KeyError):
        cls.load(save_path, adata=model.adata)


@pytest.mark.parametrize("cls", [TOTALVI])
def test_with_minified_mdata_get_latent_representation(cls):
    model, _, _, _ = prep_model_mudata(cls=cls, use_size_factor=True)

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    latent_repr_orig = model.get_latent_representation()

    model.minify_mudata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    latent_repr_new = model.get_latent_representation()

    assert_approx_equal(latent_repr_new, latent_repr_orig)


@pytest.mark.parametrize("cls", [TOTALVI])
def test_with_minified_mdata_get_feature_correlation_matrix(cls):
    model, _, _, _ = prep_model_mudata(cls=cls, use_size_factor=True)

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    fcm_orig = model.get_feature_correlation_matrix(
        correlation_type="spearman",
        transform_batch=["batch_0", "batch_1"],
    )

    model.minify_mudata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    fcm_new = model.get_feature_correlation_matrix(
        correlation_type="spearman",
        transform_batch=["batch_0", "batch_1"],
    )

    assert_approx_equal(fcm_new, fcm_orig)



from pathlib import Path

import numpy as np
import pytest

from scvi.data import synthetic_iid
from scvi.external import SCBASSET

_DNA_CODE_KEY = "code"


def _get_adata(sparse_format: str | None = None):
    dataset1 = synthetic_iid(batch_size=100, sparse_format=sparse_format).transpose()
    dataset1.X = (dataset1.X > 0).astype(float)
    dataset1.obsm[_DNA_CODE_KEY] = np.random.randint(0, 3, size=(dataset1.n_obs, 1344))
    return dataset1


def test_scbasset():
    adata = _get_adata()
    SCBASSET.setup_anndata(
        adata,
        dna_code_key=_DNA_CODE_KEY,
    )
    model = SCBASSET(adata)
    model.train(max_epochs=2, early_stopping=True)
    model.get_latent_representation()


def test_scbasset_batch():
    adata = _get_adata()
    SCBASSET.setup_anndata(
        adata,
        dna_code_key=_DNA_CODE_KEY,
        batch_key="batch",
    )
    model = SCBASSET(adata)
    model.train(max_epochs=1)
    model.get_latent_representation()
    assert hasattr(model.module, "batch_ids")


@pytest.mark.internet
def test_scbasset_motif_download(save_path):
    # get a temporary directory name
    SCBASSET._download_motifs(genome="human", motif_dir=save_path)

    assert Path(save_path, "shuffled_peaks.fasta").exists()
    assert Path(save_path, "shuffled_peaks_motifs", "MYOD1.fasta").exists()
    return


@pytest.mark.internet
def test_scbasset_tf(save_path):
    adata = _get_adata()
    SCBASSET.setup_anndata(
        adata,
        dna_code_key=_DNA_CODE_KEY,
    )
    model = SCBASSET(adata)
    # model.train(max_epochs=2, early_stopping=True)
    model.is_trained = True

    motif_seqs, bg_seqs = model._get_motif_library(tf="MYOD1", motif_dir=save_path, genome="human")

    model.get_tf_activity(tf="MYOD1", motif_dir=save_path, genome="human", lib_size_norm=True)
    model.get_tf_activity(tf="MYOD1", motif_dir=save_path, genome="human", lib_size_norm=False)
    return



import pytest

import scvi
from scvi.data import synthetic_iid
from scvi.model import SCVI, JaxSCVI
from scvi.train import JaxTrainingPlan, TrainingPlan
from scvi.train._constants import METRIC_KEYS
from scvi.train._trainingplans import _compute_kl_weight


@pytest.mark.parametrize(
    ("current", "n_warm_up", "min_kl_weight", "max_kl_weight", "expected"),
    [
        (0, 400, 0.0, 1.0, 0.0),
        (200, 400, 0.0, 1.0, 0.5),
        (400, 400, 0.0, 1.0, 1.0),
        (0, 400, 0.5, 1.0, 0.5),
        (200, 400, 0.5, 1.0, 0.75),
        (400, 400, 0.0, 2.0, 2.0),
    ],
)
def test_compute_kl_weight_linear_annealing(
    current, n_warm_up, min_kl_weight, max_kl_weight, expected
):
    kl_weight = _compute_kl_weight(current, 1, n_warm_up, None, max_kl_weight, min_kl_weight)
    assert kl_weight == pytest.approx(expected)
    kl_weight = _compute_kl_weight(1, current, None, n_warm_up, max_kl_weight, min_kl_weight)
    assert kl_weight == pytest.approx(expected)


@pytest.mark.parametrize("max_kl_weight", [1.0, 2.0])
def test_compute_kl_weight_no_annealing(max_kl_weight):
    assert _compute_kl_weight(1, 1, None, None, max_kl_weight, 0.0) == max_kl_weight


def test_compute_kl_weight_min_greater_max():
    with pytest.raises(ValueError):
        _compute_kl_weight(1, 1, 400, None, 0.5, 1.0)


@pytest.mark.parametrize(
    ("epoch", "step", "n_epochs_kl_warmup", "n_steps_kl_warmup", "expected"),
    [
        (0, 100, 100, 100, 0.0),
        (50, 200, 100, 1000, 0.5),
        (100, 200, 100, 1000, 1.0),
    ],
)
def test_compute_kl_precedence(epoch, step, n_epochs_kl_warmup, n_steps_kl_warmup, expected):
    kl_weight = _compute_kl_weight(epoch, step, n_epochs_kl_warmup, n_steps_kl_warmup, 1.0, 0.0)
    assert kl_weight == expected


def test_loss_args():
    """Test that self._loss_args is set correctly."""
    adata = synthetic_iid()
    SCVI.setup_anndata(adata)
    JaxSCVI.setup_anndata(adata)
    vae = SCVI(adata)
    jax_vae = JaxSCVI(adata)
    tp = TrainingPlan(vae.module)
    jax_tp = JaxTrainingPlan(jax_vae.module)

    loss_args = [
        "tensors",
        "inference_outputs",
        "generative_outputs",
        "kl_weight",
    ]
    assert len(tp._loss_args) == len(loss_args)
    assert len(jax_tp._loss_args) == len(loss_args)
    for arg in loss_args:
        assert arg in tp._loss_args
        assert arg in jax_tp._loss_args


def test_semisupervisedtrainingplan_metrics():
    adata = scvi.data.synthetic_iid(n_labels=3)
    scvi.model.SCANVI.setup_anndata(
        adata,
        labels_key="labels",
        unlabeled_category="label_0",
        batch_key="batch",
    )
    model = scvi.model.SCANVI(adata)
    model.train(max_epochs=1, check_val_every_n_epoch=1)

    for mode in ["train", "validation"]:
        for metric in [
            METRIC_KEYS.ACCURACY_KEY,
            METRIC_KEYS.F1_SCORE_KEY,
            METRIC_KEYS.CLASSIFICATION_LOSS_KEY,
        ]:
            assert f"{mode}_{metric}" in model.history_



import importlib
from collections.abc import Callable
from functools import wraps


def error_on_missing_dependencies(*modules):
    missing_modules = []
    for module in modules:
        try:
            importlib.import_module(module)
        except ImportError:
            missing_modules.append(module)
    if len(missing_modules) > 0:
        raise ModuleNotFoundError(f"Please install {missing_modules} to use this functionality.")


def dependencies(*modules) -> Callable:
    """Decorator to check for dependencies."""

    def decorator(fn: Callable) -> Callable:
        @wraps(fn)
        def wrapper(*args, **kwargs):
            error_on_missing_dependencies(*modules)
            return fn(*args, **kwargs)

        return wrapper

    return decorator



from typing import NamedTuple


class _MODULE_KEYS(NamedTuple):
    X_KEY: str = "x"
    # inference
    Z_KEY: str = "z"
    QZ_KEY: str = "qz"
    QZM_KEY: str = "qzm"
    QZV_KEY: str = "qzv"
    LIBRARY_KEY: str = "library"
    QL_KEY: str = "ql"
    BATCH_INDEX_KEY: str = "batch_index"
    Y_KEY: str = "y"
    CONT_COVS_KEY: str = "cont_covs"
    CAT_COVS_KEY: str = "cat_covs"
    SIZE_FACTOR_KEY: str = "size_factor"
    # generative
    PX_KEY: str = "px"
    PL_KEY: str = "pl"
    PZ_KEY: str = "pz"
    # loss
    KL_L_KEY: str = "kl_divergence_l"
    KL_Z_KEY: str = "kl_divergence_z"


MODULE_KEYS = _MODULE_KEYS()



from __future__ import annotations

import logging
import warnings
from collections import defaultdict
from functools import partial
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from typing import Literal

    from anndata import AnnData
    from mudata import MuData

    from scvi._types import Number

import numpy as np
import pandas as pd
import sparse
import torch

from scvi import REGISTRY_KEYS, settings
from scvi.data import AnnDataManager, fields
from scvi.data._constants import _SETUP_ARGS_KEY
from scvi.external.methylvi._utils import _context_cov_key, _context_mc_key
from scvi.model.base import (
    ArchesMixin,
    BaseModelClass,
    UnsupervisedTrainingMixin,
    VAEMixin,
)
from scvi.model.base._de_core import (
    _de_core,
)
from scvi.utils import setup_anndata_dsp

from ._module import METHYLVAE
from ._utils import scmc_raw_counts_properties

logger = logging.getLogger(__name__)


class METHYLVI(VAEMixin, UnsupervisedTrainingMixin, ArchesMixin, BaseModelClass):
    """
    Model class for methylVI :cite:p:`Weinberger2023a`

    Parameters
    ----------
    mdata
        MuData object that has been registered via :meth:`~scvi.external.METHYLVI.setup_mudata`.
    n_hidden
        Number of nodes per hidden layer.
    n_latent
        Dimensionality of the latent space.
    n_layers
        Number of hidden layers used for encoder and decoder NNs.
    **model_kwargs
        Keyword args for :class:`~scvi.external.methylvi.METHYLVAE`

    Examples
    --------
    >>> mdata = mudata.read_h5mu(path_to_mudata)
    >>> MethylVI.setup_mudata(mdata, batch_key="batch")
    >>> vae = MethylVI(mdata)
    >>> vae.train()
    >>> mdata.obsm["X_methylVI"] = vae.get_latent_representation()
    """

    def __init__(
        self,
        mdata: MuData,
        n_hidden: int = 128,
        n_latent: int = 10,
        n_layers: int = 1,
        **model_kwargs,
    ):
        super().__init__(mdata)

        n_batch = self.summary_stats.n_batch
        n_cats_per_cov = (
            self.adata_manager.get_state_registry(REGISTRY_KEYS.CAT_COVS_KEY)[
                fields.CategoricalJointObsField.N_CATS_PER_KEY
            ]
            if REGISTRY_KEYS.CAT_COVS_KEY in self.adata_manager.data_registry
            else None
        )

        self.contexts = self.get_anndata_manager(mdata, required=True).registry[_SETUP_ARGS_KEY][
            "methylation_contexts"
        ]
        self.num_features_per_context = [mdata[context].shape[1] for context in self.contexts]

        n_input = np.sum(self.num_features_per_context)

        self.module = METHYLVAE(
            n_input=n_input,
            n_hidden=n_hidden,
            n_latent=n_latent,
            n_layers=n_layers,
            n_batch=n_batch,
            n_cats_per_cov=n_cats_per_cov,
            contexts=self.contexts,
            num_features_per_context=self.num_features_per_context,
            **model_kwargs,
        )
        self._model_summary_string = (
            "Overwrite this attribute to get an informative representation for your model"
        )
        # necessary line to get params that will be used for saving/loading
        self.init_params_ = self._get_init_params(locals())

        logger.info("The model has been initialized")

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        **kwargs,
    ) -> AnnData | None:
        """
        %(summary)s.

        Parameters
        ----------
        %(param_adata)s

        Returns
        -------
        %(returns)s
        """
        raise NotImplementedError("METHYLVI must be used with a MuData object.")

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_mudata(
        cls,
        mdata: MuData,
        mc_layer: str,
        cov_layer: str,
        methylation_contexts: Iterable[str],
        batch_key: str | None = None,
        categorical_covariate_keys: list[str] | None = None,
        modalities=None,
        **kwargs,
    ):
        """%(summary_mdata)s.

        Parameters
        ----------
        %(param_mdata)s
        mc_layer
            Layer containing methylated cytosine counts for each set of methylation features.
        cov_layer
            Layer containing total coverage counts for each set of methylation features.
        methylation_contexts
            List of modality fields in `mdata` object representing different methylation contexts.
            Each context must be equipped with a layer containing the number of methylated counts
            (specified by `mc_layer`) and total number of counts (specified by `cov_layer`) for
            each genomic region feature.
        %(param_batch_key)s
        %(param_cat_cov_keys)s
        %(param_modalities)s

        Examples
        --------
        MethylVI.setup_mudata(
            mdata,
            mc_layer="mc",
            cov_layer="cov",
            batch_key="Platform",
            methylation_modalities=['mCG', 'mCH'],
            modalities={
                "batch_key": "mCG"
            },
        )

        """
        if modalities is None:
            modalities = {}
        setup_method_args = METHYLVI._get_setup_method_args(**locals())

        if methylation_contexts is None:
            raise ValueError("Methylation contexts cannot be None.")

        modalities_ = cls._create_modalities_attr_dict(modalities, setup_method_args)

        batch_field = fields.MuDataCategoricalObsField(
            REGISTRY_KEYS.BATCH_KEY,
            batch_key,
            mod_key=modalities_.batch_key,
        )

        cat_cov_field = fields.MuDataCategoricalJointObsField(
            REGISTRY_KEYS.CAT_COVS_KEY,
            categorical_covariate_keys,
            mod_key=modalities_.categorical_covariate_keys,
        )

        mc_fields = []
        cov_fields = []

        for context in methylation_contexts:
            mc_fields.append(
                fields.MuDataLayerField(
                    _context_mc_key(context),
                    mc_layer,
                    mod_key=context,
                    is_count_data=True,
                    mod_required=True,
                )
            )

            cov_fields.append(
                fields.MuDataLayerField(
                    _context_cov_key(context),
                    cov_layer,
                    mod_key=context,
                    is_count_data=True,
                    mod_required=True,
                )
            )

        mudata_fields = mc_fields + cov_fields + [batch_field] + [cat_cov_field]
        adata_manager = AnnDataManager(fields=mudata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(mdata, **kwargs)

        cls.register_manager(adata_manager)

    @torch.inference_mode()
    def posterior_predictive_sample(
        self,
        mdata: MuData | None = None,
        n_samples: int = 1,
        batch_size: int | None = None,
    ) -> dict[str, sparse.GCXS] | sparse.GCXS:
        r"""
        Generate observation samples from the posterior predictive distribution.

        The posterior predictive distribution is written as :math:`p(\hat{x} \mid x)`.

        Parameters
        ----------
        mdata
            MuData object with equivalent structure to initial MuData. If `None`, defaults to the
            MuData object used to initialize the model.
        n_samples
            Number of samples for each cell.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.

        Returns
        -------
        x_new : :py:class:`torch.Tensor`
            tensor with shape (n_cells, n_regions, n_samples)
        """
        mdata = self._validate_anndata(mdata)

        scdl = self._make_data_loader(adata=mdata, batch_size=batch_size)

        x_new = defaultdict(list)
        for tensors in scdl:
            samples = self.module.sample(
                tensors,
                n_samples=n_samples,
            )

            for context in self.contexts:
                x_new[context].append(sparse.GCXS.from_numpy(samples[context].numpy()))

        for context in self.contexts:
            x_new[context] = sparse.concatenate(
                x_new[context]
            )  # Shape (n_cells, n_regions, n_samples)

        return x_new

    @torch.inference_mode()
    def get_normalized_methylation(
        self,
        mdata: MuData | None = None,
        indices: Sequence[int] | None = None,
        region_list: Sequence[str] | None = None,
        n_samples: int = 1,
        n_samples_overall: int = None,
        batch_size: int | None = None,
        return_mean: bool = True,
        return_numpy: bool | None = None,
        context: str | None = None,
        **importance_weighting_kwargs,
    ) -> (np.ndarray | pd.DataFrame) | dict[str, np.ndarray | pd.DataFrame]:
        r"""Returns the normalized (decoded) methylation.

        This is denoted as :math:`\mu_n` in the methylVI paper.

        Parameters
        ----------
        mdata
            MuData object with equivalent structure to initial Mudata.
            If `None`, defaults to the MuData object used to initialize the model.
        indices
            Indices of cells in mdata to use. If `None`, all cells are used.
        region_list
            Return frequencies of expression for a subset of regions.
            This can save memory when working with large datasets and few regions are
            of interest.
        n_samples
            Number of posterior samples to use for estimation.
        n_samples_overall
            Number of posterior samples to use for estimation. Overrides `n_samples`.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        return_mean
            Whether to return the mean of the samples.
        return_numpy
            Return a :class:`~numpy.ndarray` instead of a :class:`~pandas.DataFrame`.
            DataFrame includes region names as columns. If either `n_samples=1` or
            `return_mean=True`, defaults to `False`. Otherwise, it defaults to `True`.
        context
            If not `None`, returns normalized methylation levels for the specified
            methylation context. Otherwise, a dictionary with contexts as keys and normalized
            methylation levels as values is returned.

        Returns
        -------
        If `n_samples` is provided and `return_mean` is False,
        this method returns a 3d tensor of shape (n_samples, n_cells, n_regions).
        If `n_samples` is provided and `return_mean` is True, it returns a 2d tensor
        of shape (n_cells, n_regions).
        In this case, return type is :class:`~pandas.DataFrame` unless `return_numpy` is True.
        Otherwise, the method expects `n_samples_overall` to be provided and returns a 2d tensor
        of shape (n_samples_overall, n_regions).

        If model was set up using a MuData object, a dictionary is returned with keys
        corresponding to individual methylation contexts with values determined as
        described above.
        """
        mdata = self._validate_anndata(mdata)

        if context is not None and context not in self.contexts:
            raise ValueError(
                f"{context} is not a valid methylation context for this model. "
                f"Valid contexts are {self.contexts}."
            )

        if indices is None:
            indices = np.arange(mdata.n_obs)
        if n_samples_overall is not None:
            assert n_samples == 1  # default value
            n_samples = n_samples_overall // len(indices) + 1
        scdl = self._make_data_loader(adata=mdata, indices=indices, batch_size=batch_size)

        region_mask = slice(None) if region_list is None else mdata.var_names.isin(region_list)

        if n_samples > 1 and return_mean is False:
            if return_numpy is False:
                warnings.warn(
                    "`return_numpy` must be `True` if `n_samples > 1` and `return_mean` "
                    "is`False`, returning an `np.ndarray`.",
                    UserWarning,
                    stacklevel=settings.warnings_stacklevel,
                )
            return_numpy = True

        exprs = defaultdict(list)

        for tensors in scdl:
            inference_kwargs = {"n_samples": n_samples}
            inference_outputs, generative_outputs = self.module.forward(
                tensors=tensors,
                inference_kwargs=inference_kwargs,
                generative_kwargs={},
                compute_loss=False,
            )

            for ctxt in self.contexts:
                exp_ = generative_outputs["px_mu"][ctxt]
                exp_ = exp_[..., region_mask]
                exprs[ctxt].append(exp_.cpu())

        cell_axis = 1 if n_samples > 1 else 0

        for ctxt in self.contexts:
            exprs[ctxt] = np.concatenate(exprs[ctxt], axis=cell_axis)

        if n_samples_overall is not None:
            # Converts the 3d tensor to a 2d tensor
            for ctxt in self.contexts:
                exprs[ctxt] = exprs[ctxt].reshape(-1, exprs[ctxt].shape[-1])
                n_samples_ = exprs[ctxt].shape[0]
                ind_ = np.random.choice(n_samples_, n_samples_overall, replace=True)
                exprs[ctxt] = exprs[ctxt][ind_]
                return_numpy = True

        elif n_samples > 1 and return_mean:
            for ctxt in self.contexts:
                exprs[ctxt] = exprs[ctxt].mean(0)

        if return_numpy is None or return_numpy is False:
            exprs_dfs = {}
            for ctxt in self.contexts:
                exprs_dfs[ctxt] = pd.DataFrame(
                    exprs[ctxt],
                    columns=mdata[ctxt].var_names[region_mask],
                    index=mdata[ctxt].obs_names[indices],
                )
            exprs_ = exprs_dfs
        else:
            exprs_ = exprs

        if context is not None:
            return exprs_[context]
        else:
            return exprs_

    @torch.inference_mode()
    def get_specific_normalized_methylation(
        self,
        mdata: MuData | None = None,
        context: str = None,
        indices: Sequence[int] | None = None,
        transform_batch: Sequence[Number | str] | None = None,
        region_list: Sequence[str] | None = None,
        n_samples: int = 1,
        n_samples_overall: int = None,
        weights: Literal["uniform", "importance"] | None = None,
        batch_size: int | None = None,
        return_mean: bool = True,
        return_numpy: bool | None = None,
        **importance_weighting_kwargs,
    ) -> (np.ndarray | pd.DataFrame) | dict[str, np.ndarray | pd.DataFrame]:
        r"""Convenience function to obtain normalized methylation values for a single context.

        Only applicable to MuData models.

        Parameters
        ----------
        mdata
            MuData object with equivalent structure to initial MuData. If `None`, defaults to the
            MuData object used to initialize the model.
        context
            Methylation context for which to obtain normalized methylation levels.
        indices
            Indices of cells in mdata to use. If `None`, all cells are used.
        transform_batch
            Batch to condition on.
            If transform_batch is:

            - None, then real observed batch is used.
            - int, then batch transform_batch is used.
        region_list
            Return frequencies of expression for a subset of regions.
            This can save memory when working with large datasets and few regions are
            of interest.
        n_samples
            Number of posterior samples to use for estimation.
        n_samples_overall
            Number of posterior samples to use for estimation. Overrides `n_samples`.
        weights
            Weights to use for sampling. If `None`, defaults to `"uniform"`.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        return_mean
            Whether to return the mean of the samples.
        return_numpy
            Return a :class:`~numpy.ndarray` instead of a :class:`~pandas.DataFrame`.
            DataFrame includes region names as columns. If either `n_samples=1` or
            `return_mean=True`, defaults to `False`. Otherwise, it defaults to `True`.
        importance_weighting_kwargs
            Keyword arguments passed into
            :meth:`~scvi.model.base.RNASeqMixin._get_importance_weights`.

        Returns
        -------
        If `n_samples` is provided and `return_mean` is False,
        this method returns a 3d tensor of shape (n_samples, n_cells, n_regions).
        If `n_samples` is provided and `return_mean` is True, it returns a 2d tensor
        of shape (n_cells, n_regions).
        In this case, return type is :class:`~pandas.DataFrame` unless `return_numpy` is True.
        Otherwise, the method expects `n_samples_overall` to be provided and returns a 2d tensor
        of shape (n_samples_overall, n_regions).
        """
        exprs = self.get_normalized_methylation(
            mdata=mdata,
            indices=indices,
            transform_batch=transform_batch,
            region_list=region_list,
            n_samples=n_samples,
            n_samples_overall=n_samples_overall,
            weights=weights,
            batch_size=batch_size,
            return_mean=return_mean,
            return_numpy=return_numpy,
            **importance_weighting_kwargs,
        )
        return exprs[context]

    def differential_methylation(
        self,
        mdata: MuData | None = None,
        groupby: str | None = None,
        group1: Iterable[str] | None = None,
        group2: str | None = None,
        idx1: Sequence[int] | Sequence[bool] | str | None = None,
        idx2: Sequence[int] | Sequence[bool] | str | None = None,
        mode: Literal["vanilla", "change"] = "vanilla",
        delta: float = 0.05,
        batch_size: int | None = None,
        all_stats: bool = True,
        batch_correction: bool = False,
        batchid1: Iterable[str] | None = None,
        batchid2: Iterable[str] | None = None,
        fdr_target: float = 0.05,
        silent: bool = False,
        two_sided: bool = True,
        **kwargs,
    ) -> dict[str, pd.DataFrame] | pd.DataFrame:
        r"""\.

        A unified method for differential methylation analysis.

        Implements `"vanilla"` DE :cite:p:`Lopez18`. and `"change"` mode DE :cite:p:`Boyeau19`.

        Parameters
        ----------
        %(de_mdata)s
        %(de_modality)s
        %(de_groupby)s
        %(de_group1)s
        %(de_group2)s
        %(de_idx1)s
        %(de_idx2)s
        %(de_mode)s
        %(de_delta)s
        %(de_batch_size)s
        %(de_all_stats)s
        %(de_batch_correction)s
        %(de_batchid1)s
        %(de_batchid2)s
        %(de_fdr_target)s
        %(de_silent)s
        two_sided
            Whether to perform a two-sided test, or a one-sided test.
        **kwargs
            Keyword args for :meth:`scvi.model.base.DifferentialComputation.get_bayes_factors`

        Returns
        -------
        Differential methylation DataFrame with the following columns:
        proba_de
            the probability of the region being differentially methylated
        is_de_fdr
            whether the region passes a multiple hypothesis correction procedure
            with the target_fdr threshold
        bayes_factor
            Bayes Factor indicating the level of significance of the analysis
        effect_size
            the effect size, computed as (accessibility in population 2) -
            (accessibility in population 1)
        emp_effect
            the empirical effect, based on observed detection rates instead of the estimated
            accessibility scores from the methylVI model
        scale1
            the estimated methylation level in population 1
        scale2
            the estimated methylation level in population 2
        emp_mean1
            the empirical (observed) methylation level in population 1
        emp_mean2
            the empirical (observed) methylation level in population 2

        """
        mdata = self._validate_anndata(mdata)

        def change_fn(a, b):
            return a - b

        if two_sided:

            def m1_domain_fn(samples):
                return np.abs(samples) >= delta

        else:

            def m1_domain_fn(samples):
                return samples >= delta

        result = {}
        for context in self.contexts:
            col_names = mdata[context].var_names
            model_fn = partial(
                self.get_specific_normalized_methylation,
                batch_size=batch_size,
                context=context,
            )
            all_stats_fn = partial(scmc_raw_counts_properties, context=context)

            result[context] = _de_core(
                adata_manager=self.get_anndata_manager(mdata, required=True),
                model_fn=model_fn,
                representation_fn=None,
                groupby=groupby,
                group1=group1,
                group2=group2,
                idx1=idx1,
                idx2=idx2,
                all_stats=all_stats,
                all_stats_fn=all_stats_fn,
                col_names=col_names,
                mode=mode,
                batchid1=batchid1,
                batchid2=batchid2,
                delta=delta,
                batch_correction=batch_correction,
                fdr=fdr_target,
                silent=silent,
                change_fn=change_fn,
                m1_domain_fn=m1_domain_fn,
                **kwargs,
            )

        return result



import os

import anndata
import mudata
import numpy as np
import pandas as pd
import pytest

from scvi import REGISTRY_KEYS
from scvi.data import synthetic_iid

from .utils import generic_setup_mudata_manager


def test_setup_mudata():
    adata = synthetic_iid()
    adata.obs["cont1"] = np.random.normal(size=(adata.shape[0],))
    adata.obs["cont2"] = np.random.normal(size=(adata.shape[0],))
    adata.obs["cat1"] = np.random.randint(0, 5, size=(adata.shape[0],))
    adata.obs["cat2"] = np.random.randint(0, 5, size=(adata.shape[0],))
    protein_adata = synthetic_iid()
    mdata = mudata.MuData({"rna": adata, "protein": protein_adata})
    adata_manager = generic_setup_mudata_manager(
        mdata,
        layer_mod="rna",
        layer=None,
        batch_mod="rna",
        batch_key="batch",
        categorical_covariate_mod="rna",
        categorical_covariate_keys=["cat1", "cat2"],
        continuous_covariate_mod="rna",
        continuous_covariate_keys=["cont1", "cont2"],
        protein_expression_mod="protein",
        protein_expression_layer=None,
    )
    np.testing.assert_array_equal(adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY), adata.X)
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.BATCH_KEY),
        np.array(adata.obs["_scvi_batch"]).reshape((-1, 1)),
    )
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.CONT_COVS_KEY),
        adata.obs[["cont1", "cont2"]],
    )
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.CAT_COVS_KEY),
        adata.obs[["cat1", "cat2"]],
    )
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY),
        protein_adata.X,
    )


def test_setup_mudata_view():
    # test that error is thrown if its a view:
    adata = synthetic_iid()
    mdata = mudata.MuData({"rna": adata[1]})
    with pytest.raises(ValueError):
        generic_setup_mudata_manager(mdata, layer_mod="rna")

    adata = synthetic_iid()
    mdata = mudata.MuData({"rna": adata})
    with pytest.raises(ValueError):
        generic_setup_mudata_manager(mdata[1], layer_mod="rna")


def test_setup_mudata_unpaired():
    # test that error is thrown if unpaired modalities:
    adata = synthetic_iid()
    protein_adata = synthetic_iid(batch_size=100)
    mdata = mudata.MuData({"rna": adata, "protein": protein_adata})
    with pytest.raises(ValueError):
        generic_setup_mudata_manager(mdata, layer_mod="rna", protein_expression_mod="protein")

    # Pad unpaired with zeros
    unpaired_adata = adata[mdata.obsm["rna"] & ~(mdata.obsm["protein"])]
    pad_adata = anndata.AnnData(
        pd.DataFrame(
            np.zeros((mdata.n_obs - protein_adata.n_obs, protein_adata.n_vars)),
            index=unpaired_adata.obs_names,
        )
    )
    mdata.mod["protein"] = anndata.concat([protein_adata, pad_adata], join="outer")
    mdata.update()
    generic_setup_mudata_manager(mdata, layer_mod="rna", protein_expression_mod="protein")


def test_setup_mudata_anndata():
    # test that error is thrown if an anndata is passed in:
    adata = synthetic_iid()
    with pytest.raises(AssertionError):
        generic_setup_mudata_manager(adata, layer_mod="rna")


def test_setup_mudata_layer():
    # test that layer is working properly
    adata = synthetic_iid()
    true_x = adata.X
    adata.layers["X"] = true_x
    adata.X = np.ones_like(adata.X)
    mdata = mudata.MuData({"rna": adata})
    adata_manager = generic_setup_mudata_manager(mdata, layer_mod="rna", layer="X")
    np.testing.assert_array_equal(adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY), true_x)


def test_setup_mudata_default_batch():
    # test that it creates batch if no batch_key is passed
    adata = synthetic_iid()
    mdata = mudata.MuData({"rna": adata})
    adata_manager = generic_setup_mudata_manager(mdata, layer_mod="rna", batch_mod="rna")
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.BATCH_KEY),
        np.zeros((adata.shape[0], 1)),
    )
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.BATCH_KEY),
        adata.obs["_scvi_batch"].to_numpy().reshape(-1, 1),
    )


def test_setup_mudata_nan_batch():
    # test error is thrown when categorical obs field contains nans
    adata = synthetic_iid()
    adata.obs["batch"][:10] = np.nan
    mdata = mudata.MuData({"rna": adata})
    with pytest.raises(ValueError):
        generic_setup_mudata_manager(mdata, layer_mod="rna", batch_mod="rna", batch_key="batch")


def test_save_setup_mudata(save_path):
    adata = synthetic_iid()
    mdata = mudata.MuData({"rna": adata})
    generic_setup_mudata_manager(
        mdata,
        layer_mod="rna",
        batch_mod="rna",
        batch_key="batch",
    )
    temp_path = os.path.join(save_path, "test.h5mu")
    mdata.write(temp_path)
    mudata.read(temp_path)


def test_transfer_fields():
    # test transfer_fields function
    adata1 = synthetic_iid()
    protein_adata1 = synthetic_iid()
    mdata1 = mudata.MuData({"rna": adata1, "protein": protein_adata1})
    adata2 = synthetic_iid()
    adata2.X = adata1.X
    protein_adata2 = synthetic_iid()
    mdata2 = mudata.MuData({"rna": adata2, "protein": protein_adata2})

    adata1_manager = generic_setup_mudata_manager(
        mdata1,
        layer_mod="rna",
        layer=None,
        batch_mod="rna",
        batch_key="batch",
        protein_expression_mod="protein",
        protein_expression_layer=None,
    )
    adata1_manager.transfer_fields(mdata2)
    np.testing.assert_array_equal(adata1.obs["_scvi_batch"], adata2.obs["_scvi_batch"])


def test_transfer_fields_view():
    # test transfer_fields function
    adata1 = synthetic_iid()
    protein_adata1 = synthetic_iid()
    mdata1 = mudata.MuData({"rna": adata1, "protein": protein_adata1})
    adata2 = synthetic_iid()
    adata2.X = adata1.X
    protein_adata2 = synthetic_iid()
    mdata2 = mudata.MuData({"rna": adata2, "protein": protein_adata2})[:50]

    adata1_manager = generic_setup_mudata_manager(
        mdata1,
        layer_mod="rna",
        layer=None,
        batch_mod="rna",
        batch_key="batch",
        protein_expression_mod="protein",
        protein_expression_layer=None,
    )
    with pytest.raises(ValueError):
        adata1_manager.transfer_fields(mdata2)


def test_transfer_fields_layer():
    # test if layer was used initially, again used in transfer setup
    adata1 = synthetic_iid()
    raw_counts = adata1.X.copy()
    adata1.layers["raw"] = raw_counts
    protein_adata1 = synthetic_iid()
    mdata1 = mudata.MuData({"rna": adata1, "protein": protein_adata1})
    adata2 = synthetic_iid()
    adata2.layers["raw"] = raw_counts
    protein_adata2 = synthetic_iid()
    mdata2 = mudata.MuData({"rna": adata2, "protein": protein_adata2})

    zeros = np.zeros_like(adata1.X)
    ones = np.ones_like(adata1.X)
    adata1.X = zeros
    adata2.X = ones

    adata_manager = generic_setup_mudata_manager(
        mdata1,
        layer_mod="rna",
        layer="raw",
        batch_mod="rna",
        batch_key="batch",
        protein_expression_mod="protein",
        protein_expression_layer=None,
    )
    adata_manager.transfer_fields(mdata2)
    np.testing.assert_array_equal(adata1.obs["_scvi_batch"], adata2.obs["_scvi_batch"])


def test_transfer_fields_unknown_batch():
    # test that an unknown batch throws an error
    adata1 = synthetic_iid()
    protein_adata1 = synthetic_iid()
    mdata1 = mudata.MuData({"rna": adata1, "protein": protein_adata1})
    adata2 = synthetic_iid()
    adata2.X = adata1.X
    protein_adata2 = synthetic_iid()
    mdata2 = mudata.MuData({"rna": adata2, "protein": protein_adata2})
    adata2.obs["batch"] = [2] * adata2.n_obs
    adata1_manager = generic_setup_mudata_manager(
        mdata1, layer_mod="rna", batch_mod="rna", batch_key="batch"
    )
    with pytest.raises(ValueError):
        adata1_manager.transfer_fields(mdata2)


def test_transfer_fields_diff_batch_mapping():
    # test that correct mapping was applied
    adata1 = synthetic_iid()
    protein_adata1 = synthetic_iid()
    mdata1 = mudata.MuData({"rna": adata1, "protein": protein_adata1})
    adata2 = synthetic_iid()
    adata2.X = adata1.X
    protein_adata2 = synthetic_iid()
    mdata2 = mudata.MuData({"rna": adata2, "protein": protein_adata2})
    adata2.obs["batch"] = ["batch_1"] * adata2.n_obs
    adata1_manager = generic_setup_mudata_manager(
        mdata1, layer_mod="rna", batch_mod="rna", batch_key="batch"
    )
    adata1_manager.transfer_fields(mdata2)
    batch_mapping = adata1_manager.get_state_registry(REGISTRY_KEYS.BATCH_KEY).categorical_mapping
    print(batch_mapping)
    correct_batch = np.where(batch_mapping == "batch_1")[0][0]
    assert adata2.obs["_scvi_batch"][0] == correct_batch


def test_transfer_fields_missing_batch():
    # test that transfer_fields correctly looks for adata.obs['batch']
    adata1 = synthetic_iid()
    protein_adata1 = synthetic_iid()
    mdata1 = mudata.MuData({"rna": adata1, "protein": protein_adata1})
    adata2 = synthetic_iid()
    del adata2.obs["batch"]
    adata2.X = adata1.X
    protein_adata2 = synthetic_iid()
    mdata2 = mudata.MuData({"rna": adata2, "protein": protein_adata2})
    adata1_manager = generic_setup_mudata_manager(
        mdata1, layer_mod="rna", batch_mod="rna", batch_key="batch"
    )
    with pytest.raises(KeyError):
        adata1_manager.transfer_fields(mdata2)


def test_transfer_fields_default_batch():
    # test that transfer_fields assigns same batch to cells
    # if the original anndata was also same batch
    adata1 = synthetic_iid()
    protein_adata1 = synthetic_iid()
    mdata1 = mudata.MuData({"rna": adata1, "protein": protein_adata1})
    adata2 = synthetic_iid()
    del adata2.obs["batch"]
    adata2.X = adata1.X
    protein_adata2 = synthetic_iid()
    mdata2 = mudata.MuData({"rna": adata2, "protein": protein_adata2})
    adata1_manager = generic_setup_mudata_manager(
        mdata1, layer_mod="rna", batch_mod="rna", batch_key=None
    )
    adata1_manager.transfer_fields(mdata2)
    assert adata2.obs["_scvi_batch"][0] == 0


def test_transfer_fields_covariates():
    adata1 = synthetic_iid()
    adata1.obs["cont1"] = np.random.normal(size=(adata1.shape[0],))
    adata1.obs["cont2"] = np.random.normal(size=(adata1.shape[0],))
    adata1.obs["cat1"] = np.random.randint(0, 5, size=(adata1.shape[0],))
    adata1.obs["cat2"] = np.random.randint(0, 5, size=(adata1.shape[0],))
    mdata1 = mudata.MuData({"rna": adata1})
    adata_manager = generic_setup_mudata_manager(
        mdata1,
        layer_mod="rna",
        batch_mod="rna",
        continuous_covariate_mod="rna",
        continuous_covariate_keys=["cont1", "cont2"],
        categorical_covariate_mod="rna",
        categorical_covariate_keys=["cat1", "cat2"],
    )
    adata2 = synthetic_iid()
    adata2.obs["cont1"] = np.random.normal(size=(adata2.shape[0],))
    adata2.obs["cont2"] = np.random.normal(size=(adata2.shape[0],))
    adata2.obs["cat1"] = 0
    adata2.obs["cat2"] = 1
    mdata2 = mudata.MuData({"rna": adata2})

    adata_manager.transfer_fields(mdata2)

    # give it a new category
    adata2.obs["cat1"] = 6
    adata_manager2 = adata_manager.transfer_fields(mdata2, extend_categories=True)
    assert adata_manager2.get_state_registry(REGISTRY_KEYS.CAT_COVS_KEY).mappings["cat1"][-1] == 6


def test_data_format():
    # if data was dense np array, check after setup_anndata, data is C_CONTIGUOUS
    adata = synthetic_iid()
    protein_adata = synthetic_iid()
    mdata = mudata.MuData({"rna": adata, "protein": protein_adata})

    old_x = adata.X
    old_pro = protein_adata.X
    old_obs = adata.obs
    adata.X = np.asfortranarray(old_x)
    protein_adata.X = np.asfortranarray(old_pro)
    assert adata.X.flags["C_CONTIGUOUS"] is False
    assert protein_adata.X.flags["C_CONTIGUOUS"] is False

    adata_manager = generic_setup_mudata_manager(
        mdata, layer_mod="rna", protein_expression_mod="protein"
    )
    assert adata.X.flags["C_CONTIGUOUS"] is True
    assert protein_adata.X.flags["C_CONTIGUOUS"] is True

    assert np.array_equal(old_x, adata.X)
    assert np.array_equal(old_pro, protein_adata.X)
    assert np.array_equal(old_obs, adata.obs)

    assert np.array_equal(adata.X, adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY))
    assert np.array_equal(
        protein_adata.X,
        adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY),
    )

    # if obsm is dataframe, make it C_CONTIGUOUS if it isnt
    adata = synthetic_iid()
    protein_adata = synthetic_iid()
    mdata = mudata.MuData({"rna": adata, "protein": protein_adata})

    pe = np.asfortranarray(protein_adata.X)
    protein_adata.X = pe
    assert protein_adata.X.flags["C_CONTIGUOUS"] is False

    adata_manager = generic_setup_mudata_manager(
        mdata, layer_mod="rna", protein_expression_mod="protein"
    )
    new_pe = adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY)
    assert new_pe.flags["C_CONTIGUOUS"] is True
    assert np.array_equal(pe, new_pe)
    assert np.array_equal(adata.X, adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY))
    assert np.array_equal(
        protein_adata.X,
        adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY),
    )
    assert adata.X.flags["C_CONTIGUOUS"] is True
    assert protein_adata.X.flags["C_CONTIGUOUS"] is True


def test_view_registry():
    adata = synthetic_iid()
    protein_adata = synthetic_iid()
    mdata = mudata.MuData({"rna": adata, "protein": protein_adata})
    adata_manager = generic_setup_mudata_manager(
        mdata,
        layer_mod="rna",
        layer=None,
        batch_mod="rna",
        batch_key="batch",
        protein_expression_mod="protein",
        protein_expression_layer=None,
    )
    adata_manager.view_registry()
    adata_manager.view_registry(hide_state_registries=True)



import numpy as np
import pytest

from scvi.data import synthetic_iid
from scvi.external import SOLO
from scvi.model import SCVI


@pytest.mark.parametrize("soft", [True, False])
@pytest.mark.parametrize("return_logits", [True, False])
def test_solo(soft: bool, return_logits: bool):
    n_latent = 5
    adata = synthetic_iid()
    SCVI.setup_anndata(adata)
    model = SCVI(adata, n_latent=n_latent)
    model.train(1, check_val_every_n_epoch=1, train_size=0.5)

    solo = SOLO.from_scvi_model(model)
    solo.train(1, check_val_every_n_epoch=1, train_size=0.9)
    assert "validation_loss" in solo.history.keys()
    predictions = solo.predict(soft=soft, return_logits=return_logits)
    if soft:
        preds = predictions.to_numpy()
        assert preds.shape == (adata.n_obs, 2)
        if not return_logits:
            assert np.allclose(preds.sum(axis=-1), 1)
        else:
            assert not np.allclose(preds.sum(axis=-1), 1)

    bdata = synthetic_iid()
    solo = SOLO.from_scvi_model(model, bdata)
    solo.train(1, check_val_every_n_epoch=1, train_size=0.9)
    assert "validation_loss" in solo.history.keys()
    predictions = solo.predict(soft=soft)
    if soft:
        preds = predictions.to_numpy()
        assert preds.shape == (adata.n_obs, 2)
        assert np.allclose(preds.sum(axis=-1), 1)


def test_solo_multiple_batch():
    n_latent = 5
    adata = synthetic_iid()
    adata.layers["my_layer"] = adata.X.copy()
    SCVI.setup_anndata(adata, layer="my_layer", batch_key="batch")
    model = SCVI(adata, n_latent=n_latent)
    model.train(1, check_val_every_n_epoch=1, train_size=0.5)

    solo = SOLO.from_scvi_model(model, restrict_to_batch="batch_0")
    solo.train(1, check_val_every_n_epoch=1, train_size=0.9)
    assert "validation_loss" in solo.history.keys()
    solo.predict()


def test_solo_scvi_labels():
    n_latent = 5
    adata = synthetic_iid()
    SCVI.setup_anndata(adata, labels_key="labels")
    model = SCVI(adata, n_latent=n_latent)
    model.train(1, check_val_every_n_epoch=1, train_size=0.5)

    solo = SOLO.from_scvi_model(model)
    solo.train(1, check_val_every_n_epoch=1, train_size=0.9)
    assert "validation_loss" in solo.history.keys()
    solo.predict()


def test_solo_from_scvi_errors():
    adata = synthetic_iid()
    adata.obs["continuous_covariate"] = np.random.normal(size=(adata.n_obs, 1))
    adata.obs["categorical_covariate"] = np.random.choice(["a", "b", "c"], size=(adata.n_obs, 1))

    # no batch key, restrict_to_batch
    SCVI.setup_anndata(adata, labels_key="labels")
    model = SCVI(adata)
    model.train(max_epochs=1)
    with pytest.raises(ValueError):
        _ = SOLO.from_scvi_model(model, restrict_to_batch="batch_0")

    # continuous covariate
    SCVI.setup_anndata(
        adata,
        batch_key="batch",
        labels_key="labels",
        continuous_covariate_keys=["continuous_covariate"],
    )
    model = SCVI(adata)
    model.train(max_epochs=1)

    with pytest.raises(ValueError):
        _ = SOLO.from_scvi_model(model, restrict_to_batch="batch_0")

    # categorical covariate
    SCVI.setup_anndata(
        adata,
        batch_key="batch",
        labels_key="labels",
        categorical_covariate_keys=["categorical_covariate"],
    )
    model = SCVI(adata)
    model.train(max_epochs=1)

    with pytest.raises(ValueError):
        _ = SOLO.from_scvi_model(model, restrict_to_batch="batch_0")



import numpy as np

from scvi.data import synthetic_iid
from scvi.external import RNAStereoscope, SpatialStereoscope


def test_stereoscope(save_path):
    dataset = synthetic_iid(
        n_labels=5,
    )
    RNAStereoscope.setup_anndata(
        dataset,
        labels_key="labels",
    )

    # train with no proportions
    sc_model = RNAStereoscope(dataset)
    sc_model.train(max_epochs=1)

    # train again with proportions
    sc_model = RNAStereoscope(dataset, ct_weights=np.ones((5,)))
    sc_model.train(max_epochs=1)
    # test save/load
    sc_model.save(save_path, overwrite=True, save_anndata=True)
    sc_model = RNAStereoscope.load(save_path)
    dataset = synthetic_iid(
        n_labels=5,
    )
    SpatialStereoscope.setup_anndata(
        dataset,
    )
    st_model = SpatialStereoscope.from_rna_model(dataset, sc_model, prior_weight="minibatch")
    st_model.train(max_epochs=1)
    st_model.get_proportions()
    # test save/load
    st_model.save(save_path, overwrite=True, save_anndata=True)
    st_model = SpatialStereoscope.load(save_path)
    st_model.get_proportions()

    # try imputation code
    y = np.array(50 * ["label_0"])
    st_model.get_scale_for_ct(y)
